"""
TUI Screen (purdy.tui.iscreen.py)
=================================

This module is an Urwid code viewer concrete implementaion of a Screen. It is
constructed by :class:`purdy.ui.Screen` depending on its factory.
"""
import urwid

from purdy.colour import COLOURIZERS
from purdy.iscreen.tui.widgets import (CodeWidget, DividingLine,
    TwinContainer, ScrollingListBox, ScrollingIndicator)


UrwidColourizer = COLOURIZERS['urwid']

#import logging
#logging.basicConfig(filename='debug.log', level=logging.DEBUG)
#logger = logging.getLogger()

# =============================================================================
# Window Management
# =============================================================================

class FocusWalker:
    ### Urwid has no call for focusing on a widget, you have to use an index
    # into its parent's container. The FocusWalker finds all of the CodeWidget
    # objects, figures out which one is currently focused and allows for the
    # previous and next item to be focused
    class Node:
        def __init__(self, position, parent=None, parent_position=0):
            self.position = position
            self.parent = parent
            self.parent_position = parent_position

    def __init__(self, root):
        ### Figures out the order things should be focused in. 
        #
        # :param root: root container to look for CodeWidgets within
        #
        self.root = root
        self.focused_index = 0
        self.nodes = []

        current_focus, _ = root.contents[root.focus_position]

        # root is a Pile with either CodeWidgets or Columns in it, where
        # the columns contain CodeWidgets
        for position, content in enumerate(root.contents):
            widget = content[0]
            if isinstance(widget, CodeWidget):
                if widget == current_focus:
                    self.focused_index = len(self.nodes)

                self.nodes.append( self.Node(position) )
            elif isinstance(widget, urwid.Columns):
                index = len(self.nodes)

                self.nodes.append( self.Node(position, widget, 0) )
                self.nodes.append( self.Node(position, widget, 1) )

                if widget == current_focus:
                    if widget.focus_position == 0:
                        self.focused_index = index
                    elif widget.focus_position == 1:
                        self.focused_index = index + 1
                    # else: shouldn't happen

            # else: shouldn't happen

    def focus_next(self):
        # if we're the last item, loop to the next
        position = self.focused_index + 1
        if position >= len(self.nodes):
            position = 0

        self.set_focus(position)

    def focus_prev(self):
        # if we're the first item, loop to the next
        position = self.focused_index - 1
        if position < 0:
            position = len(self.nodes) - 1

        self.set_focus(position)

    def set_focus(self, position):
        # set the focus
        node = self.nodes[position]
        self.root.focus_position = node.position

        if node.parent:
            node.parent.focus_position = node.parent_position


class BaseWindow(urwid.Pile):
    def __init__(self, iscreen, *args, **kwargs):
        self.iscreen = iscreen
        self.parent_screen = iscreen.parent_screen
        self.animation_manager = iscreen.parent_screen.animation_manager
        super(BaseWindow, self).__init__(*args, **kwargs)

    def _next_focus(self):
        walker = FocusWalker(self)
        walker.focus_next()

    def _prev_focus(self):
        walker = FocusWalker(self)
        walker.focus_prev()

    def _close_help(self):
        self.iscreen.loop.widget = self

    def _show_help(self):
        help_box = HelpDialog(self)
        self.iscreen.loop.widget = urwid.Overlay(help_box, self, 
            align='center', valign='middle', width=55, height=20)

    def keypress(self, size, key):
        if key in ('q', 'Q'):
            raise urwid.ExitMainLoop()

        if key == '?':
            self._show_help()
            return

        if key in ('f1', 'f2', 'f3', 'f4', 'f5', 'f6', 'f7', 'f8', 'f9', 'f10'):
            # let parent handle function keys
            return super(BaseWindow, self).keypress(size, key)

        if key not in self.animation_manager.handled_keys: 
            # handle keys not handled by the animation manager 
            result = super(BaseWindow, self).keypress(size, key)
            if result is None:
                # keypress was handled by child, we're done
                return None
            # else: fall through if super().keypress() didn't deal with it

        if self.parent_screen.movie_mode > -1:
            # Ignore other keypresses in movie mode
            return None

        # --- at this point the keypress was not handled by the children, see
        # if we want to do anything with it
        if key == 'tab':
            self._next_focus()
            return None

        if key == 'shift tab':
            self._prev_focus()

        return self.animation_manager.perform(key)

    def animation_alarm(self, loop, data):
        self.animation_manager.animation_alarm()

    def movie_alarm(self, loop, data):
        self.animation_manager.movie_alarm()

    def auto_forward_alarm(self, loop, data):
        self.animation_manager.auto_forward_alarm()

# =============================================================================
# Screen
# =============================================================================

class ConcreteCodeBox:
    """:class:`purdy.ui.CodeBox` represents a box of code in the 
    :class:`purdy.ui.Screen`. This is an Urwid implementation of it. 

    :param proxy_code_box: the :class:`purdy.ui.CodeBox` representing what is 
                           to be built.
    """
    def __init__(self, proxy_code_box):
        self.proxy_code_box = proxy_code_box
        self.widget = CodeWidget(self, self.proxy_code_box.auto_scroll)

    def build(self, container):
        self.proxy_code_box.listing.set_display('urwid', self.widget)

        if self.proxy_code_box.height == 0:
            container.append(self.widget)
        else:
            # height was specified, return a tuple instead
            container.append( (self.proxy_code_box.height, self.widget) )


class ConcreteTwinCodeBox:
    """:class:`purdy.ui.TwinCodeBox` represents two boxes of code in the 
    :class:`purdy.ui.Screen`, this is an Urwid implementation of it.

    :param proxy: the :class:`purdy.ui.TwinCodeBox` representing what is to be 
                  built.

    """
    def __init__(self, proxy):
        class Side:
            pass

        self.proxy = proxy
        self.left = Side()
        self.left.weight = proxy.left_weight
        self.left.box = ConcreteCodeBox(proxy.left.code_box)

        self.right = Side()
        self.right.weight = proxy.right_weight
        self.right.box = ConcreteCodeBox(proxy.right.code_box)

    def build(self, container):
        # cant' use CodeBox's build() method because we need to pack them into 
        # Urwid # differently
        self.proxy.left.code_box.listing.set_display('urwid', 
            self.left.box.widget)
        self.proxy.right.code_box.listing.set_display('urwid', 
            self.right.box.widget)

        # Combine the widgets from the CodeBoxes into a display container
        render_left = ('weight', self.left.weight, self.left.box.widget)
        render_right = ('weight', self.right.weight, self.right.box.widget)

        twin = TwinContainer([render_left, render_right], dividechars=1)

        if self.proxy.height != 0:
            twin = (self.proxy.height, twin)

        container.append(twin)


class TUIScreen:
    """Concrete, Urwid based implementation of a screen.

    :param parent_screen: :class:`purdy.ui.Screen` object that is creating
                          this concrete implementation

    """
    def __init__(self, parent_screen):
        self.parent_screen = parent_screen
        self.widgets = []

    def _add_concrete_code_box(self, code_box, constructor):
        size = len(self.widgets)
        if size > 0:
            # not the first box, add a dividing line
            compact = self.parent_screen.code_boxes[size - 1].compact
            div_height = 3
            if compact:
                div_height = 1

            divider = (div_height, DividingLine())
            self.widgets.append(divider)

        box = constructor(code_box)
        box.build(self.widgets)

    def add_code_box(self, code_box):
        self._add_concrete_code_box(code_box, ConcreteCodeBox)

    def add_twin_code_box(self, code_box):
        self._add_concrete_code_box(code_box, ConcreteTwinCodeBox)

    def set_alarm(self, handler, when):
        fn = getattr(self.base_window, handler)
        alarm_handle = self.loop.set_alarm_in(when, fn)
        return alarm_handle

    def remove_alarm(self, alarm_handle):
        self.loop.remove_alarm(alarm_handle)

    def run(self):
        """Calls the main display event loop. Does not return until the UI
        exits."""
        self.base_window = BaseWindow(self, self.widgets)
        window = self.base_window
        if self.parent_screen.max_height:
            # Force a maximum height on the window, put the BaseWindow in a
            # container and fill anything that is bigger than max
            window = urwid.Pile([
                (int(self.parent_screen.max_height), self.base_window),
                urwid.SolidFill(' '),
            ])

        palette = UrwidColourizer.create_urwid_palette()
        self.loop = urwid.MainLoop(window, palette)

        if self.parent_screen.settings['colour'] == 256:
            # don't confuse urwid's screen with ours
            self.loop.screen.set_terminal_properties(colors=256)
            self.loop.screen.reset_default_terminal_palette()

        # as soon as the loop is going invoke the first animation
        self.loop.set_alarm_in(0, self.base_window.auto_forward_alarm)

        # call urwid's main loop, this code doesn't return until the loop
        # exits!!!
        self.loop.run()

# =============================================================================
# Help Dialog
# =============================================================================

class HelpDialog(urwid.Pile):
    def __init__(self, parent):
        self.parent = parent

        ### Define dialog
        header = urwid.Text(('reverse', 'Help'), align = 'center')

        # Body
        index = parent.animation_manager.index + 1
        steps = len(parent.animation_manager.cells)

        walker = urwid.SimpleListWalker([
            urwid.Text(''),
            urwid.Text(('title', 'Keys')),
            urwid.Text(''),
            urwid.Text([('bold', 'q, Q   '), ': Quit']),
            urwid.Text([('bold', '<TAB>  '), ': Switch focused window']),
            urwid.Text([('bold', '⬆ <TAB>'), ': Previous focused window']),
            urwid.Text([('bold', '<LEFT> '), ': Previous step']),
            urwid.Text([('bold', '<RIGHT>'), ': Next step']),
            urwid.Text([('bold', 's      '), ': Next step, skip animation']),
            urwid.Text([('bold', 'S      '), ': Skip to next section marker']),
            urwid.Text(               '         or to the end'),
            urwid.Text(''),
            urwid.Text(('title', 'Multiples')),
            urwid.Text('You can perform multiple s, S, and <LEFT> commands'),
            urwid.Text('by typing a number before the command'),
            urwid.Text(''),
            urwid.Text(('title', 'State')),
            urwid.Text(f'Step {index} of {steps}'),
        ])

        scroller = ScrollingIndicator()
        listbox = ScrollingListBox(scroller, walker)
        layout = [listbox, (1, scroller)]
        body = urwid.Columns(layout)

        # Footer
        footer = urwid.Text(('bold', "<ENTER> to close"), align="center")

        # Layout
        layout = urwid.Frame(body, header, footer, 'body')
        line_box = urwid.LineBox(layout)
        
        super().__init__([line_box])

    def keypress(self, size, key):
        if key == 'enter':
            self.parent._close_help()

        super().keypress(size, key)



from typing import (Iterable, Iterator, Union, List, Callable, Optional, Any,
    Tuple)
from purdy.actions import (AppendTypewriter, Append, Suffix, Wait, Clear,
    Insert, Replace, Remove, Shell, InsertTypewriter, SuffixTypewriter,
    Highlight, HighlightChain, Fold, StopMovie, Transition, Sleep, RunFunction,
    Section)
from purdy.content import Code
from purdy.ui import VirtualCodeBox, Screen


class ActionsBuilder(Iterable):
    """Collects all actions to be called on a :class:`purdy.ui.Screen`.
    Once all actions are created the instance of this class can be passed
    to the :class:`purdy.ui.Screen.run()` method.

    :param screen: screen containing the code boxes the actions should be
                   added to

    :param lexer_name: name of the lexer to be used, which must be one of the
                       following:

                        - con: Python 3 Console
                        - py3: Python 3 Source
                        - bash, Bash Console

    Example
    =======

    >>> from purdy.ui import SimpleScreen
    >>> screen = SimpleScreen(starting_line_number=-1)
    >>> actions = (
    ...     ActionsBuilder(code_box, "bash")
    ...     .append_typewriter(text="$ ls ~/.pyenv/versions/")
    ...     .append(text="2.7.15 3.9.0 3.9.1")
    ... )
    >>> screen.run(actions)
    """

    def __init__(self, screen: Screen, lexer_name: str) -> None:
        self.__screen = screen
        self.__code_box = self.__screen.code_boxes[0]
        self.__lexer_name = lexer_name
        self.__actions = []

    def __iter__(self) -> Iterator[Any]:
        return iter(self.__actions)

    def switch_to_code_box(self, index: int) -> "ActionsBuilder":
        """Switches to another code box on screen so the following actions
        will be executed there

        :param index: index of the code box where the following actions
        should be executed
        :return: self
        """
        self.__code_box = self.__screen.code_boxes[index]
        return self

    def _add_action(self, action: Any) -> "ActionsBuilder":
        self.__actions.append(action)
        return self

    def _create_code(self, filename: str = "", text: str = "", 
            code: Code = None) -> Optional[Code]:

        if filename or text or code:
            return code or Code(filename=filename, text=text, 
                lexer_name=self.__lexer_name)

    def insert(self, position: int, filename: str = "", text: str = "", 
            code: Code = None) -> "ActionsBuilder":
        """Adds an :class:`purdy.actions.Insert` action

        :param position: line number to insert code at. Position is 1-indexed.
                         Content is pushed down, so a value of "1" inserts at 
                         the beginning. Negative indices are supported. A
                         value of "0" will append the code to the bottom.

        :param filename: name of a file to read for :class:`purdy.content.Code` 
                         content.  If both this and `text` is given,
                         `filename` is used first

        :param text: text to read for :class:`purdy.content.Code` content.

        :param code: a :class:`purdy.content.Code` object containing the source
                     code to insert. If all `code`, `filename` and `text` are 
                     given, `code` will be used first.

        :return: self
        """
        return self._add_action(Insert(self.__code_box, position, 
            self._create_code(filename, text, code)))

    def append(self, filename: str = "", text: str = "", 
            code: Code = None) -> "ActionsBuilder":
        """Adds an :class:`purdy.actions.Append` action

        :param filename: name of a file to read for :class:`purdy.content.Code` 
                         content.  If both this and `text` is given,
                         `filename` is used first

        :param text: text to read for :class:`purdy.content.Code` content.

        :param code: a :class:`purdy.content.Code` object containing the source
                     code to insert. If all `code`, `filename` and `text` are 
                     given, `code` will be used first.

        :return: self
        """
        return self._add_action(Append(self.__code_box,
                self._create_code(filename, text, code)))

    def replace(self, position: int, filename: str = "", text: str = "", 
            code: Code = None) -> "ActionsBuilder":
        """Adds an :class:`purdy.actions.Replace` action

        :param position: line number to replace the code at. Position is 
                         1-indexed.  Negative indices are supported.

        :param filename: name of a file to read for :class:`purdy.content.Code` 
                         content. If both this and `text` is given, `filename`
                         is used first

        :param text: text to read for :class:`purdy.content.Code` content.

        :param code: a :class:`purdy.content.Code` object containing the source
                     code to insert. If all `code`, `filename` and `text` are 
                     given, `code` will be used first.

        :return: self
        """
        return self._add_action(Replace(self.__code_box, position, 
            self._create_code(filename, text, code)))

    def remove(self, position: int, size: int) -> "ActionsBuilder":
        """Adds an :class:`purdy.actions.Remove` action

        :param position: line number to replace the code at. Position is 
                         1-indexed.  Negative indices are supported.

        :param size: number of lines to remove.

        :return: self
        """
        return self._add_action(Remove(self.__code_box, position, size))

    def clear(self) -> "ActionsBuilder":
        """Adds an :class:`purdy.actions.Clear` action

        :return: self
        """
        return self._add_action(Clear(self.__code_box))

    def suffix(self, position: int, source: str) -> "ActionsBuilder":
        """Adds an :class:`purdy.actions.Suffix` action

        :param position: line number to replace the code at. Position is 
                         1-indexed.  Negative indices are supported.

        :param source: string containing content to append to the line

        :return: self
        """
        return self._add_action(Suffix(self.__code_box, position, source))

    def shell(self, cmd: str) -> "ActionsBuilder":
        """Adds an :class:`purdy.actions.Shell` action

        :param cmd: string containing the shell command and its parameters.
                    Example: ``ls -la``.

        :return: self
        """
        return self._add_action(Shell(self.__code_box, cmd))

    def append_typewriter(self, filename: str = "", text: str = "", 
            code: Code = None) -> "ActionsBuilder":
        """Adds an :class:`purdy.actions.AppendTypewriter` action

        :param filename: name of a file to read for :class:`purdy.content.Code` 
                         content.  If both this and `text` is given,
                         `filename` is used first

        :param text: text to read for :class:`purdy.content.Code` content.

        :param code: a :class:`purdy.content.Code` object containing the source
                     code to insert. If all `code`, `filename` and `text` are
                     given, `code` will be used first.

        :return: self
        """
        return self._add_action(AppendTypewriter(self.__code_box, 
            self._create_code(filename, text, code)))

    def insert_typewriter(self, position: int, filename: str = "", 
            text: str = "", code: Code = None) -> "ActionsBuilder":
        """Adds an :class:`purdy.actions.InsertTypewriter` action

        :param position: line number to insert the code at. Position is 
                         1-indexed.

        :param filename: name of a file to read for :class:`purdy.content.Code` 
                         content. If both this and `text` is given, `filename`
                         is used first

        :param text: text to read for :class:`purdy.content.Code` content.

        :param code: a :class:`purdy.content.Code` object containing the source
                     code to insert. If all `code`, `filename` and `text` are 
                     given, `code` will be used first.

        :return: self
        """
        return self._add_action(InsertTypewriter(self.__code_box, position, 
            self._create_code(filename, text, code)))

    def suffix_typewriter(self, position: int, source: str) -> "ActionsBuilder":
        """Adds an :class:`purdy.actions.SuffixTypewriter` action

        :param position: line number to insert the code at. Position is 
                         1-indexed.  Negative indices are supported.

        :param source: a string to be appended to the given line

        :return: self
        """
        return self._add_action(SuffixTypewriter(self.__code_box, position, 
            source))

    def highlight(self, spec: Union[int, str, Iterable[int]], 
            highlight_on: bool) -> "ActionsBuilder":
        """Adds an :class:`purdy.actions.Highlight` action

        :param spec: either a string containing comma separated and/or hyphen
                     separated integers (e.g. "1,3,7-9") or a list of integers
                     specifying the lines in the code box to highlight. Line
                     numbers are 1-indexed

        :param highlight_on: True to turn highlighting on, False to turn it off

        :return: self
        """
        return self._add_action(Highlight(self.__code_box, spec, highlight_on))

    def highlight_chain(self, 
            spec_list: List[Union[str, List[int]]]) -> "ActionsBuilder":
        """Adds an :class:`purdy.actions.HighlightChain` action

        :param spec_list: a list of highlight specs (see :class:`Highlight` for
                          details on a spec)

        :return: self
        """
        return self._add_action(HighlightChain(self.__code_box, spec_list))

    def fold(self, position: int, end=-1) -> "ActionsBuilder":
        """Adds an :class:`purdy.actions.Fold` action

        :param position: line number to begin the fold at. Position is 
                         1-indexed.

        :param end: line number to finish the folding at, inclusive. A value of
                    -1 can be used to fold to the end of the box. Defaults to
                    -1.

        :return: self
        """
        return self._add_action(Fold(self.__code_box, position, end))

    def wait(self) -> "ActionsBuilder":
        """Adds an :class:`purdy.actions.Wait` action

        :return: self
        """
        return self._add_action(Wait())

    def stop_movie(self) -> "ActionsBuilder":
        """Adds an :class:`purdy.actions.StopMovie` action

        :return: self
        """
        return self._add_action(StopMovie())

    def transition(self, filename: str = "", text: str = "", code: Code = None,
        code_box_to_copy: VirtualCodeBox = None) -> "ActionsBuilder":
        """Adds an :class:`purdy.actions.Transition` action

        :param filename: name of a file to read for :class:`purdy.content.Code` 
                         content. If both this and `text` is given, `filename`
                         is used first

        :param text: text to read for :class:`purdy.content.Code` content.

        :param code: a :class:`purdy.content.Code` object containing the source
                     code to insert. If all `code`, `filename` and `text` are 
                     given, `code` will be used first.

        :param code_box_to_copy: a code box containing rendered code to copy 
                                 into this one to display. This is typically a
                                 :class:`VirtualCodeBox`. Should not be used
                                 at the same time as code parameter.

        :return: self
        """
        return self._add_action(Transition(self.__code_box,
                self._create_code(filename, text, code), code_box_to_copy))

    def sleep(self, 
            time: Union[float, Tuple[float, float]]) -> "ActionsBuilder":
        """Adds an :class:`purdy.actions.Sleep` action

        :param time: amount of time to sleep given in seconds, or a tuple
                     specifying lower and upper bound of random time in seconds 

        :return: self
        """
        return self._add_action(Sleep(time))

    def run_function(self, fn: Callable, 
            undo: Optional[Callable], *args, **kwargs) -> "ActionsBuilder":
        """Adds an :class:`purdy.actions.RunFunction` action

        :param fn: function to be called
        :param undo: function to be called when this Action is undone
        :param *args, **kwargs: any remaining arguments are passed to the
                                functions when they are called

        :return: self
        """
        return self._add_action(RunFunction(fn, undo, *args, **kwargs))

    def section(self) -> 'ActionsBuilder':
        """Adds a :class:`purdy.actions.Section` action

        :return: self
        """
        return self._add_action(Section())



"""
Wigets (purdy.wigets.py)
========================

Widgets for displaying. These are called and managed through the Screen
classes in :mod:`purdy.ui`.

"""
import urwid

# =============================================================================
# Widgets
# =============================================================================

class DividingLine(urwid.Filler):
    tab_focusable = False

    def __init__(self):
        divider = urwid.Divider('-')

        super(DividingLine, self).__init__(divider)

# -----------------------------------------------------------------------------
# CodeWidget -- box that displays code

class ScrollingIndicator(urwid.Frame):
    def __init__(self):
        self.up = urwid.Text(' ')
        self.down = urwid.Text(' ')

        # create this Frame with a solid fill in the middle and the up and
        # down Text widgets as the header and footer
        super(ScrollingIndicator, self).__init__(urwid.SolidFill(' '), 
            header=self.up, footer=self.down)

    def set_up(self, is_up, focus):
        if is_up and focus:
            self.up.set_text('▲')
        elif is_up and not focus:
            self.up.set_text('△')
        else:
            self.down.set_text(' ')

    def set_down(self, is_down, focus):
        if is_down and focus:
            self.down.set_text('▼')
        elif is_down and not focus:
            self.down.set_text('▽')
        else:
            self.down.set_text(' ')

    def set_focus_only(self, focus):
        if focus: 
            self.up.set_text('▮')
        else: 
            self.up.set_text('▯')


class ScrollingListBox(urwid.ListBox):
    def __init__(self, scroll_indicator, *args, **kwargs):
        self.scroll_indicator = scroll_indicator
        super(ScrollingListBox, self).__init__(*args, **kwargs)

    def render(self, size, focus):
        result = super(ScrollingListBox, self).render(size, focus)

        # ends_visible() returns a list with the words "top" and/or "bottom"
        # in it if the top and/or bottom of the list box is visible
        #
        # set our scrolling indicators based on what is visible
        visible = self.ends_visible(size)
        if 'top' in visible and 'bottom' in visible:
            self.scroll_indicator.set_focus_only(focus)
        else:
            self.scroll_indicator.set_up('top' not in visible, focus)
            self.scroll_indicator.set_down('bottom' not in visible, focus)

        return result

    def keypress(self, size, key):
        if key == 'up' and self.focus_position == 0:
            # don't want arrow up to change parent's focus, eat this key
            return None

        if key == 'down' and self.focus_position + 1 >= len(self.body):
            # don't want arrow down to change parent's focus, eat this key
            return None

        return super().keypress(size, key)


class CodeWidget(urwid.Columns):
    """Urwid widget that displays the code. This implements the methods of
    :class:`purdy.content.RenderHook` and is registered against a
    :class:`purdy.ui.CodeBox` and :class:`purdy.content.Listing`. As changes
    are made to the listing they will be rendered this widget.

    The widget wraps an urwid ListBox, with each line in the box being a line
    of code. It also provides indiciators on the right side of the screen as
    to whether there is content above or below the current screen. If the
    parent :class:`Screen` implementation has multiple instances of this class
    active, the scroll area will also indicate which code box is focused. 

    The up and down arrows as well as the page-up and page-down buttons are
    supported. If there are multiple code widgets, tab key will change the
    focus.
    """
    tab_focusable = True

    # CodeWidget is a Column collection containing a ListBox of Text for the
    # code and a side bar with indicators about focus and scroll position
    def __init__(self, screen, auto_scroll):
        """Constructor. These objects should only be constructed by a parent
        :class:`Screen` object.

        :param screen: the :class:`Screen` building this code box
        """
        self.screen = screen
        self.auto_scroll = auto_scroll

        # urwid does weird things when trying to focus an empty listing, never
        # allow it to be empty
        self.is_empty = True
        self.walker = urwid.SimpleListWalker([urwid.Text(''), ])

        scroller = ScrollingIndicator()
        self.listbox = ScrollingListBox(scroller, self.walker)

        layout = [self.listbox, (1, scroller)]

        super(CodeWidget, self).__init__(layout)

    #--- RenderHook methods
    def line_inserted(self, listing, position, line):
        markup = listing.render_line(line)
        index = position - 1

        if self.is_empty:
            self.walker.contents[0] = urwid.Text(markup)
            self.is_empty = False
        else:
            self.walker.contents.insert(index, urwid.Text(markup))

        if self.auto_scroll:
            # if auto scrolling, change focus to last inserted item
            self.listbox.set_focus(index)

    def line_removed(self, listing, position):
        if len(self.walker.contents) == 1:
            self.is_empty = True
            self.walker.contents[0] = urwid.Text('')
        else:
            del self.walker.contents[position - 1]

        # urwid crashes if the focus is set outside of the range and you
        # try to do other operations to the box before returning to the
        # event loop, fix this by setting the focus to the last item
        size = len(self.walker.contents)
        if size > 0:
            self.listbox.set_focus(size - 1)

    def line_changed(self, listing, position, line):
        markup = listing.render_line(line)
        index = position - 1
        self.walker.contents[index].set_text(markup)

    def clear(self):
        self.is_empty = True
        self.walker.contents.clear()
        self.walker.contents.insert(0, urwid.Text(''))


class TwinContainer(urwid.Columns):
    ### A column with the tab_focusable attribute so the BaseWindow class
    # knows to do tab focus changes on it
    tab_focusable = True



from pygments.token import Keyword, Name, Comment, String, Error, \
    Number, Operator, Generic, Token, Whitespace, Punctuation

from purdy.parser import FoldedCodeLine, token_ancestor

# =============================================================================
# RTF Colourizer

import struct

def rtf_encode(text):
    """RTF uses a weird UTF-16 decimal escape sequence for encoding unicode.
    This method take a string and returns an RTF encoded version of it."""
    output = []
    data = text.encode('utf-16le')
    length = len(data) // 2

    # use struct.unpack to turn the pairs of bytes into decimal unicode escape
    # numbers
    parts = struct.unpack(f'<{length}H', data)
    index = 0
    size = len(parts)

    while index < size:
        num = parts[index]

        if num <= 127:
            # regular ascii
            letter = chr(num)

            # RTF uses backslash, so it needs to be escaped
            if letter == '\\':
                output.append('\\\\')
            else:
                output.append(letter)
        elif num <= 256:
            # extended ascii, use hex notation
            output.append(f"\\'{num:2x}" )
        elif num < 55296:
            # 0xD800 (55296) is the the boundary for two-word encoding, less
            # than this means it is a single item
            output.append(f"\\uc0\\u{num}")
        else:
            # greater than 0xD800, means two words
            index += 1
            next_num = parts[index]
            output.append(f"\\uc0\\u{num} \\u{next_num}")

        index += 1

    return ''.join(output)

# -----------------------------------------------------------------------------

class RTFHeaders:
    preamble = r"""{\rtf1\ansi\ansicpg1252
{\fonttbl\f0\fnil\fcharset0 RobotoMono-Regular;}
"""
    color_table = r"""
{\colortbl;\red255\green255\blue255;\red143\green89\blue2;\red32\green72\blue167;\red206\green92\blue0;
\red0\green0\blue0;\red52\green101\blue164;\red204\green0\blue0;\red92\green53\blue204;\red78\green154\blue6;
\red0\green0\blue207;\red239\green41\blue41;\red164\green0\blue0;
\red213\green213\blue213;}
"""

    color_table_background_template = r"""
{\colortbl;\red255\green255\blue255;\red143\green89\blue2;\red32\green72\blue167;\red206\green92\blue0;
\red0\green0\blue0;\red52\green101\blue164;\red204\green0\blue0;\red92\green53\blue204;\red78\green154\blue6;
\red0\green0\blue207;\red239\green41\blue41;\red164\green0\blue0;
\red213\green213\blue213;%s}
"""

    font_spec = r'\f0\fs28' + '\n'

# -----------------------------------------------------------------------------

_code_palette = {
    Token:              '\\cf0 %s\n',
    Whitespace:         '\\cf0 %s\n',
    Comment:            '\\i \\cf2 %s\n\\i0\n',
    Keyword:            '\\b \\cf3 %s\n\\b0\n',
    Operator:           '\\b \\cf4 %s\n\\b0\n', 
    Name:               '\\cf5 %s\n',
    Name.Builtin:       '\\cf6 %s\n', 
    Name.Function:      '\\cf5 %s\n',
    Name.Class:         '\\cf5 %s\n',
    Name.Exception:     '\\b \\cf7 %s\n\\b0\n', 
    Name.Decorator:     '\\b \\cf8 %s\n\\b0\n', 
    Name.Variable:      '\\cf0 %s\n',
    Name.Constant:      '\\cf0 %s\n',
    Name.Attribute:     '\\cf0 %s\n',
    Name.Tag:           '\\cf0 %s\n',
    Punctuation:        '\\b \\cf5 %s\n\\b0\n', 
    String:             '\\cf9 %s\n', 
    String.Doc:         '\\i \\cf2 %s\n \\i0', 
    Number:             '\\b \cf10 %s\n \\b0', 
    Generic.Prompt:     '\\cf2 %s\n',
    Generic.Error:      '\\cf11 %s\n',
    Generic.Traceback:  '\\b \cf12 %s\n\\b0\n', 
    Error:              '\\cf11 %s\n',
}

_xml_palette = dict(_code_palette)
_xml_palette.update({
    Name.Attribute: '\\b \\cf3 %s\n\\b0\n',
    Name.Tag:       '\\cf6 %s\n', 
    Keyword:        '\\cf6 %s\n', 
    Punctuation:    '\\cf6 %s\n', 
})

class RTFColourizer:
    palettes = {
        'code':_code_palette,
        'xml':_xml_palette,
    }

    rtf_header = RTFHeaders.preamble + RTFHeaders.color_table + \
        RTFHeaders.font_spec

    highlight = '\\cb13 %s \\cb1'

    @classmethod
    def set_background_colour(cls, colour):
        """RTF doesn't really support turning background colours off, once you
        have a background colour it can only set the whole doc to something
        else. This makes highlighting weird if your background colour isn't
        white. Calling this method updates the colour table with the specified
        value and changes the document to have that background colour

        :param colour: string containing a HTML hex colour code without the
                       starting #, e.g. "ccc", "f3b2e1"
        """
        double_it = lambda s, i: f'0x{s[i]}{s[i]}'

        # calculate the integer value of the hex colour
        if len(colour) == 3:
            red = int(double_it(colour, 0), base=16)
            green = int(double_it(colour, 1), base=16)
            blue = int(double_it(colour, 2), base=16)
        else:
            red = int(f'0x{colour[0:2]}', base=16)
            green = int(f'0x{colour[2:4]}', base=16)
            blue = int(f'0x{colour[4:6]}', base=16)

        bg = f'\\red{red}\\green{green}\\blue{blue};'  # rtf colour descriptor

        # update the header to include the new background colour and the
        # highlight spec to use it
        cls.rtf_header = RTFHeaders.preamble + \
            RTFHeaders.color_table_background_template % bg + \
            RTFHeaders.font_spec + '\\cb14 \n'
        cls.highlight = '\\cb13 %s \\cb14'

    @classmethod
    def reset_background_colour(cls):
        """When testing, need to put the class back in its original state."""
        cls.rtf_header = RTFHeaders.preamble + RTFHeaders.color_table + \
            RTFHeaders.font_spec

    @classmethod
    def colourize(cls, code_line):
        """Returns a text string with RTF style tags. RTF uses a colour table,
        this styling assumes you're using the same colour table, available in
        the header in :attr:`RTFColourizer.rtf_header`.

        :param code_line: a :class:`CodeLine`   object to colourize
        """
        if isinstance(code_line, FoldedCodeLine):
            return '\\cf0 ⋮'

        palette = code_line.lexer.palette
        ancestor_list = cls.palettes[palette].keys()

        output = []
        if code_line.line_number >= 0:
            output.append( cls.line_number(code_line.line_number) )

        for part in code_line.parts:
            text = rtf_encode(part.text)
            key = token_ancestor(part.token,   ancestor_list)
            render = cls.palettes[palette][key] % text

            if code_line.highlight:
                render = cls.highlight % render

            output.append(render)

        return ''.join(output)

    @classmethod
    def line_number(cls, num):
        """Returns a colourized version of a line number"""
        return f'\\cf5 {num:3} \n'



from pathlib import Path

from waelstow import capture_stdout

from purdy.content import Code, Listing
from purdy.colour import COLOURIZERS

from purdy.scribe import (range_set_to_list, print_tokens, print_ansi,
    print_html, print_rtf)

from tests.base import PurdyContentTest

RTFColourizer = COLOURIZERS['rtf']

# =============================================================================

class TestScribe(PurdyContentTest):
    def test_range_set_to_list(self):
        expected = [3,]
        result = range_set_to_list('3')
        self.assertEqual(expected, result)

        expected = [3, 4, 5]
        result = range_set_to_list('3-5')
        self.assertEqual(expected, result)

        expected = [3, 4, 5, 6, 7, 8, 9]
        result = range_set_to_list('3-5,9,4,5-8')
        self.assertEqual(expected, result)

    def file_compare(self, datafile, fn, *args):
        with capture_stdout() as captured:
            fn(*args)

            filename = Path(__file__).parent / f'data/{datafile}'
            with open(filename.resolve()) as f:
                expected = f.read()

        self.assertEqual(expected, captured.getvalue())

    def test_tokens(self):
        code = Code(text=self.py_source, lexer_name='py3')
        listing = Listing(code)

        self.file_compare('print_tokens_ansi.out', print_tokens, listing, True)
        self.file_compare('print_tokens_bw.out', print_tokens, listing, False)

    def test_ansi(self):
        code = Code(text=self.py_source, lexer_name='py3')
        listing = Listing(code)
        listing.set_display('ansi')
        self.file_compare('print_ansi.out', print_ansi, listing)

    def test_html(self):
        code = Code(text=self.py_source, lexer_name='py3')
        listing = Listing(code)
        listing.set_display('html')
        self.file_compare('print_html_snippet.out', print_html, listing, True)
        self.file_compare('print_html_full.out', print_html, listing, False)

    def test_rtf(self):
        RTFColourizer.reset_background_colour()
        code = Code(text=self.py_source, lexer_name='py3')
        listing = Listing(code)
        listing.set_display('rtf')
        self.file_compare('print_rtf_nobg.out', print_rtf, listing)
        self.file_compare('print_rtf_bg.out', print_rtf, listing, 'ccc')

# Uncomment to regenerate a data file
#    def test_generate(self):
#        print('*** Generating data files')
#
#        ### This stub code is used to generate data files to compare against
#        code = Code(text=self.py_source, lexer_name='py3')
#        listing = Listing(code)
#
#        config = {
#            #'print_tokens_ansi.out':[
#            #    None,
#            #    print_tokens, 
#            #    [True, ],
#            #],
#            #'print_tokens_bw.out':[
#            #    None,
#            #    print_tokens, 
#            #    [False, ],
#            #],
#            #'print_ansi.out':[
#            #    'ansi',
#            #    print_ansi, 
#            #    [],
#            #],
#            #'print_rtf_nobg.out':[
#            #    'rtf',
#            #    print_rtf, 
#            #    [],
#            #],
#            #'print_rtf_bg.out':[
#            #    'rtf',
#            #    print_rtf, 
#            #    ['ccc', ],
#            #],
#            #'print_html_snippet.out':[
#            #    'html',
#            #    print_html, 
#            #    [True],
#            #],
#            #'print_html_full.out':[
#            #    'html',
#            #    print_html, 
#            #    [False],
#            #],
#        }
#
#        for filename, parms in config.items():
#            print('    => generating', filename)
#            with capture_stdout() as captured:
#                if parms[0]:
#                    listing.set_display(parms[0])
#
#                fn = parms[1]
#                fn(listing, *parms[2])
#
#                with open(filename, 'w') as f:
#                    f.write(captured.getvalue())



import inspect
from unittest import TestCase

from purdy.builder import ActionsBuilder
from purdy import actions

class TestBuilderCoverage(TestCase):
    def test_builder_coverage(self):
        ### Checks the ActionsBuilder has a method for each of the actions in
        # the actions module

        ignore_functions = ['init', 'iter', 'createcode', 'addaction', 
            'switchtocodebox', 'new', 'subclasshook']
        ignore_actions = ['codepart', 'typewriterstep', 'typewriterbase',
            'codeline', ]

        # create a list of the named actions by introspecting the actions
        # module, removing anything in the ignore list
        action_names = set([name.lower() for name, _ in \
            inspect.getmembers(actions, inspect.isclass)])
        action_names.difference_update(set(ignore_actions))

        # build a list of the functions in the ActionsBuilder, removing
        # anything in the ignore list
        function_names = set([name.replace('_', '') for name, _ in \
            inspect.getmembers(ActionsBuilder, inspect.isfunction)])
        function_names.difference_update(set(ignore_functions))

        self.assertEqual(action_names, function_names)



#!/usr/bin/env python
import os, sys, importlib, json
from pathlib import Path

# =============================================================================

def load_samples_module(mod_name):
    samples = Path(__file__).resolve()
    samples = samples.parent.parent / 'extras/samples'
    samples = samples.resolve()
    mod_name = mod_name.replace('.', '/')
    filename = samples / f'{mod_name}.py'

    # Change to samples dir, most of the scripts have local referenced files, 
    # then save so we can reset afterwards
    current_dir = os.getcwd()
    os.chdir(samples)

    # Load the file as if it were a module
    spec = importlib.util.spec_from_file_location(mod_name, filename)
    module = importlib.util.module_from_spec(spec)
    sys.modules[mod_name] = module
    spec.loader.exec_module(module)

    os.chdir(current_dir)

    return module


def run_exercise(module_name):
    from purdy.settings import settings
    settings['deactivate_args'] = True

    module = load_samples_module(module_name)

    # Set up to use the ExerciseScreen implementation
    from purdy.ui import Factory
    Factory.name = 'exercise'

    if Factory.iscreen:
        # This has been called before in the same session (Factory is a
        # singleton), clear out the flipbook so it is empty
        Factory.iscreen.flipbook.pages = []

    module.screen.run(module.actions)

    return module.screen.iscreen.flipbook.pages


def capture(module_name):
    pages = run_exercise(module_name)

    with open('flipbook.json', 'w') as f:
        json.dump(pages, f, indent=2)


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(description=('Captures the Cells used in '
        'the named sample program. Uses the ExerciseScreen mechanism that '
        'captures each change in a series of flip book pages. '
        'File must use global variables "screen" and "actions".'))

    parser.add_argument('module_name', help=('Name of module to load. Expects '
        'corresponding <module_name>.py file in the extras/samples directory'))

    args = parser.parse_args()
    capture(args.module_name)



"""
Colour Module (purdy.colour)
===============================

Contains classes to convert tokens to colour according to the various
supported renderers and palettes.
"""

from . import plainco, ansico, htmlco, rtfco, urwidco

COLOURIZERS = {
    'plain':plainco.PlainColourizer,
    'ansi':ansico.ANSIColourizer,
    'html':htmlco.HTMLColourizer,
    'rtf':rtfco.RTFColourizer,
    'urwid':urwidco.UrwidColourizer,
}



"""
Scribe Module (purdy.scribe.py)
===============================

Methods for transforming code into different representations on stdout.
"""
from colored import fg, bg, attr

from purdy.colour import COLOURIZERS

RTFColourizer = COLOURIZERS['rtf']

# =============================================================================
# Utility Methods
# =============================================================================

def range_set_to_list(text):
    ### converts a range set of numbers to a sorted list of unique numbers
    #
    # e.g: 3-5,9,8,4 => [3, 4, 5, 8, 9]
    values = set()
    parts = text.split(',')
    for part in parts:
        if '-' in part:
            start, stop = part.split('-')
            start = int(start)
            stop = int(stop) + 1
            values.update(range(start, stop))
        else:
            values.add(int(part))

    return sorted(values)

# =============================================================================
# Scribe to <stdout> Methods
# =============================================================================

def print_tokens(listing, colour=True):
    """Prints each line in a :class:`purdy.content.Code` object with a
    coloured background, then prints the parsed tokens inside that line

    :param listing: :class:`purdy.content.Listing` object containing code to
                    print

    :param colour: set to True to print out using ANSI colour. Defaults to True
    """
    for line in listing.lines:
        text = line.text
        if text == '':
            text = '\\n'

        if colour:
            print(fg('white') + attr('bold') + bg('blue') + text +attr('reset'))
        else:
            print(text)

        for part in line.parts:
            output = part.text
            if output == '':
                output = "''"

            if colour:
                if output != "''":
                    output = fg('black') + bg('grey_70') + output +attr('reset')

                print(f'    {part.token}: {output}')
            else:
                if output == '':
                    output = "''"
                print(f'    {part.token}: {output}')

# -----------------------------------------------------------------------------

def print_ansi(listing):
    for row in listing.content():
        print(row)

# -----------------------------------------------------------------------------

def print_html(listing, snippet=True):
    """Prints the code in an HTML representation.

    :param listing: :class:`Listing` object containing code to print
    :param snippet: if True, prints out just the <div> containing the code.
                    Otherwise, prints a full valid HTML file. Defaults to
                    True.
    """
    output = []
    if not snippet:
        output.extend([
            '<!doctype html>',
            '<html lang="en">',
            '<body>',
        ])

    output.append( ('<div style="background :#f8f8f8; overflow:auto; '
        'width:auto; border:solid gray; border-width:.1em .1em .1em .8em; '
        'padding:.2em .6em;"><pre style="margin: 0; line-height:125%">')
    )

    for row in listing.content():
        output.append( row + '\n')

    output.append('</pre></div>\n')

    if not snippet:
        output.append('</body>\n</html>')
    print(''.join(output))

# -----------------------------------------------------------------------------

def print_rtf(listing, background_colour=None):
    """Prints an RTF document containing the colourized code

    :param listing: :class:`Listing` object containing code to print
    """
    if background_colour:
        listing.colourizer.set_background_colour(background_colour)

    output = [listing.colourizer.rtf_header, ]

    for row in listing.content():
        output.append(row)
        output.append('\\\n')   # \ is rtf newline

    output.append('}')
    print(''.join(output))



"""
Exercise Screen (purdy.Exercise.iscreen.py)
===========================================

This module mimics a code viewer, capturing the picture of every code box in
the screen at every step. Used for testing actions and steps.
"""

# =============================================================================
# Screen
# =============================================================================

class FlipBook:
    def __init__(self):
        self.pages = []

    def add_page(self, code_box, change_type, content):
        data = {
            'code_box':f'CodeBox(id={code_box.id})',
            'change_type':change_type,
            'content':content,
        }
        self.pages.append(data)


class ExerciseCodeWidget:
    """Render hook implementation that stores actions as flip books.
    """
    def __init__(self, flipbook, code_book_count):
        self.lines = []
        self.flipbook = flipbook

        self.id = code_book_count

    #--- RenderHook methods
    def line_inserted(self, listing, position, line):
        content = listing.render_line(line)
        index = position - 1
        self.lines.insert(index, content)
        
        self.flipbook.add_page(self, f'insert:{position}', content)

    def line_removed(self, listing, position):
        del self.lines[position - 1]

        self.flipbook.add_page(self, f'remove:{position}', '')

    def line_changed(self, listing, position, line):
        content = listing.render_line(line)
        index = position - 1
        self.lines[index] = content

        self.flipbook.add_page(self, f'change:{position}', content)

    def clear(self):
        self.lines = []

        self.flipbook.add_page(self, 'clear', '')


class ExerciseScreen:
    """Concrete implementation of a screen that captures each change to its
    codeboxes in "flip book pages". Each page can be thought of a step in the
    animations used to create the pages. Nothing is displayed, this is used
    for testing.

    :param parent_screen: :class:`purdy.ui.Screen` object that is creating
                          this concrete implementation
    """
    def __init__(self, parent_screen):
        self.parent_screen = parent_screen
        self.widgets = {}
        self.flipbook = FlipBook()
        self.mode = 'plain'
        self.code_book_count = 0

    def add_code_box(self, code_box):
        self.code_book_count += 1
        widget = ExerciseCodeWidget(self.flipbook, self.code_book_count)
        self.widgets[code_box] = widget
        code_box.listing.set_display(self.mode, widget)

    def add_twin_code_box(self, twin_code_box):
        self.code_book_count += 1
        widget = ExerciseCodeWidget(self.flipbook, self.code_book_count)
        self.widgets[twin_code_box.left.code_box] = widget
        twin_code_box.left.code_box.listing.set_display(self.mode, widget)

        self.code_book_count += 1
        widget = ExerciseCodeWidget(self.flipbook, self.code_book_count)
        self.widgets[twin_code_box.right.code_box] = widget
        twin_code_box.right.code_box.listing.set_display(self.mode, widget)

    def set_alarm(self, handler, when):
        pass

    def remove_alarm(self, alarm_handle):
        pass

    def run(self):
        """Runs the actions on the code listings."""
        manager = self.parent_screen.animation_manager
        # Run through the steps forward, then backwards then forward again
        self.flipbook.pages.append('=== Actions ===')
        for action in self.parent_screen.actions:
            self.flipbook.pages.append(str(action))

        self.flipbook.pages.append('=== Steps ===')
        for cell in manager.cells:
            for step in cell.steps:
                self.flipbook.pages.append(str(step))

        self.flipbook.pages.append('=== Forward ===')
        for cell in manager.cells:
            cell.render(manager, skip=True)

        self.flipbook.pages.append('=== Backward ===')
        for cell in reversed(manager.cells):
            cell.undo(manager)

        self.flipbook.pages.append('=== Forward ===')
        for cell in manager.cells:
            cell.render(manager, skip=True)



"""
Command Line Tools
==================

Several of the command line tools have common arguments and needs. This file 
defines helper functions so these are defined once.
"""
from purdy.__init__ import __version__
from purdy.parser import PurdyLexer

LEXER_HELP = ('Name of lexer to use to parse the file. Choices are: ' 
    f'{PurdyLexer.choices()}. If no choice given, attempts to determine the '
    'result automatically. If it cannot detect it, it assumes Python 3.')

# =============================================================================
# Argument Builders for argparse
# =============================================================================

def max_height_arg(parser):
    parser.add_argument('--maxheight', help=('Sets a maximum screen height '
        'for the TUI screen viewer. Ignored if not in TUI mode.'), type=int, 
        default=0)

def num_arg(parser):
    parser.add_argument('--num', help=('Display line numbers with code '
        'starting with the value given here'), type=int, default=-1)


def lexer_arg(parser):
    parser.add_argument('-l', '--lexer', choices=PurdyLexer.names(),
        default='detect', help=LEXER_HELP)


def version_arg(parser):
    parser.add_argument('--version', action='version', 
        version='%(prog)s {version}'.format(version=__version__ ))


def filename_arg(parser):
    parser.add_argument('filename', help='Name of file to parse')


def highlight_arg(parser):
    parser.add_argument('--highlight', '--hl', help=('Highlight certain line '
        'numbers when displaying the code. Line numbers are 1-indexed. '
        'Multiple lines can be higlighted using a hyphen for range (e.g. 1-4, '
        'inclusive) or a comma separated list (e.g. 1-4,7,9 is line 1, 2, 3, '
        '4, 7 and 9).'))


def background_arg(parser):
    parser.add_argument('--background', '--bg', help=('Change the background '
        'colour of the document. When using the --highlight option, you '
        'you should also set a background colour, otherwise the background '
        'will turn white due to how RTF supports colouring. Format of the '
        'colour is like an HTML colour, e.g. #c1b455, without the leading '
        '#'))

# =============================================================================
# Meta Argument Builders
# =============================================================================

def general_args(parser):
    lexer_arg(parser)
    version_arg(parser)


def rtf_args(parser):
    num_arg(parser)
    background_arg(parser)
    highlight_arg(parser)


def ansi_args(parser):
    num_arg(parser)
    highlight_arg(parser)



"""
Virtual Screen (purdy.virtual.iscreen.py)
=========================================

This module mimics a code viewer, running through the requested actions and
making the final result available.
"""

from purdy.colour import COLOURIZERS
from purdy.ui import VirtualCodeWidget

RTFColourizer = COLOURIZERS['rtf']

# =============================================================================
# Screen
# =============================================================================

class VirtualScreen:
    """Concrete, Urwid based implementation of a screen.

    :param parent_screen: :class:`purdy.ui.Screen` object that is creating
                          this concrete implementation

    """
    def __init__(self, parent_screen):
        self.parent_screen = parent_screen
        self.widgets = {}
        self.mode = 'plain'
        if parent_screen.args.exportrtf:
            self.mode = 'rtf'

        self.background_colour = parent_screen.args.background

    def add_code_box(self, code_box):
        widget = VirtualCodeWidget()
        self.widgets[code_box] = widget
        code_box.listing.set_display(self.mode, widget)

    def add_twin_code_box(self, twin_code_box):
        widget = VirtualCodeWidget()
        self.widgets[twin_code_box.left.code_box] = widget
        twin_code_box.left.code_box.listing.set_display(self.mode, widget)

        widget = VirtualCodeWidget()
        self.widgets[twin_code_box.right.code_box] = widget
        twin_code_box.right.code_box.listing.set_display(self.mode, widget)

    def run(self):
        """Runs the actions on the code listings."""
        manager = self.parent_screen.animation_manager
        for cell in manager.cells:
            cell.render(manager, skip=True)

        if self.mode == 'rtf':
            if self.background_colour:
                RTFColourizer.set_background_colour(self.background_colour)

            print(RTFColourizer.rtf_header)

            for count, box in enumerate(self.parent_screen.code_boxes):
                widget = self.widgets[box]

                if count > 0:
                    print('\\\n')
                    print('\cf5 ' + 55*'=')
                    print('\\\n')
                    print('\\\n')

                for line in widget.lines:
                    print(line)
                    print('\\\n')

            print('}')
            return

        # else, plain mode
        for count, box in enumerate(self.parent_screen.code_boxes):
            widget = self.widgets[box]

            if count > 0:
                print(55*'=')

            for line in widget.lines:
                print(line)



from purdy.parser import FoldedCodeLine

# =============================================================================
# Plain Colourizer: the colourizer that doesn't do colour, handles plain text
# augmentation like line numbers for uncolourized display

class PlainColourizer:
    @classmethod
    def colourize(cls, code_line):
        """Returns the plain text version of the code line.

        :param code_line: a :class:`CodeLine` object to process
        """
        if isinstance(code_line, FoldedCodeLine):
            return '⋮'

        output = []
        if code_line.line_number != -1:
            output.append( cls.line_number(code_line.line_number) )

        output.extend([part.text for part in code_line.parts])
        return ''.join(output)

    @classmethod
    def line_number(cls, num):
        """Returns a colourized version of a line number"""
        return f'{num:3} '



from pygments.token import Keyword, Name, Comment, String, Error, \
    Number, Operator, Generic, Token, Whitespace, Punctuation

from purdy.parser import FoldedCodeLine, token_ancestor

# =============================================================================
# HTML Colourizer: augments code with colour specified using HTML style tags

import html

_code_palette = {
    Token:              '',
    Whitespace:         '',
    Comment:            'color: #8f5902; font-style: italic;',
    Keyword:            'color: #2048a7; font-weight: bold;',
    Operator:           'color: #ce5c00; font-weight: bold;',
    Name:               'color: #000000;',
    Name.Builtin:       'color: #3465a4;',
    Name.Function:      'color: #000000;',
    Name.Class:         'color: #000000;',
    Name.Exception:     'color: #cc0000; font-weight: bold;',
    Name.Decorator:     'color: #5c35cc; font-weight: bold;',
    Name.Variable:      '',
    Name.Constant:      '',
    Name.Attribute:     '',
    Name.Tag:           '',
    Punctuation:        'color: #000000; font-weight: bold;',
    String:             'color: #4e9a06;',
    String.Doc:         'color: #8f5902; font-style: italic;',
    Number:             'color: #0000cf; font-weight: bold;',
    Generic.Prompt:     'color: #8f5902;',
    Generic.Error:      'color: #ef2929;',
    Generic.Traceback:  'color: #a40000; font-weight: bold;',
    Error:              'color: #ef2929;',
}

_xml_palette = dict(_code_palette)
_xml_palette.update({
    Name.Attribute: 'color: #2048a7; font-weight: bold;',
    Name.Tag:       'color: #3465a4;',
    Keyword:        'color: #3465a4;',
    Punctuation:    'color: #3465a4;',
})

class HTMLColourizer:
    palettes = {
        'code':_code_palette,
        'xml':_xml_palette,
    }

    highlight = 'background-color:#ddd; '

    @classmethod
    def colourize(cls, code_line):
        """Returns a text string with ANSI colour escapes corresponding to the
        tokens sent in

        :param code_line: a :class:`CodeLine` object to colourize
        """
        if isinstance(code_line, FoldedCodeLine):
            return '<span style="color: #000; font-weight: bold;">⋮</span>'

        palette = code_line.lexer.palette
        ancestor_list = cls.palettes[palette].keys()

        output = []

        if code_line.line_number >= 0:
            output.append( HTMLColourizer.line_number(code_line.line_number) )
        
        if code_line.highlight:
            output.append(f'<span style="{cls.highlight}">')

        for part in code_line.parts:
            text = html.escape(part.text)
            key = token_ancestor(part.token,   ancestor_list)
            style = cls.palettes[palette][key]

            if style:
                output.append(f'<span style="{style}">{text}</span>')
            else:
                output.append(text)

        if code_line.highlight:
            output.append('</span>')

        return ''.join(output)

    @classmethod
    def line_number(cls, num):
        """Returns a colourized version of a line number"""
        style = ('color: #333; display: inline-block; min-width: 4ex;'
            'text-align: right; padding-right: 1ex;')
        return f'<span style="{style}">{num}</span>'




import importlib, sys
from importlib.machinery import SourceFileLoader

from pathlib import Path
from unittest import TestCase

from waelstow import capture_stdout

from purdy.colour import COLOURIZERS

RTFColourizer = COLOURIZERS['rtf']

# =============================================================================

PAT_EXPECTED ="""\
[38;5;241m 10[0m [38;5;18m[48;5;237m>>> [0m[38;5;62m[1m[48;5;237m@decorated[0m
[38;5;241m 11[0m [38;5;18m[48;5;237m... [0m[38;5;3m[48;5;237mdef[0m[48;5;237m [0m[38;5;7m[48;5;237mthing[0m[48;5;237m([0m[38;5;7m[48;5;237ma[0m[48;5;237m,[0m[48;5;237m [0m[38;5;7m[48;5;237mb[0m[48;5;237m,[0m[48;5;237m [0m[38;5;7m[48;5;237mc[0m[48;5;237m)[0m[48;5;237m:[0m
[38;5;241m 12[0m [38;5;18m... [0m    [0m[38;5;3mpass[0m
"""

# Careful with this string, there are necessary trailing blanks on some lines!
PRAT_EXPECTED = """{\\rtf1\\ansi\\ansicpg1252
{\\fonttbl\\f0\\fnil\\fcharset0 RobotoMono-Regular;}

{\\colortbl;\\red255\\green255\\blue255;\\red143\\green89\\blue2;\\red32\\green72\\blue167;\\red206\\green92\\blue0;
\\red0\\green0\\blue0;\\red52\\green101\\blue164;\\red204\\green0\\blue0;\\red92\\green53\\blue204;\\red78\\green154\\blue6;
\\red0\\green0\\blue207;\\red239\\green41\\blue41;\\red164\\green0\\blue0;
\\red213\\green213\\blue213;\\red204\\green170\\blue34;}
\\f0\\fs28
\\cb14 
\\cf5  10 
\\cb13 \\cf2 >>> 
 \\cb14\\cb13 \\b \\cf8 @decorated
\\b0
 \\cb14\\
\\cf5  11 
\\cb13 \\cf2 ... 
 \\cb14\\cb13 \\b \\cf3 def
\\b0
 \\cb14\\cb13 \\cf0  
 \\cb14\\cb13 \\cf5 thing
 \\cb14\\cb13 \\b \\cf5 (
\\b0
 \\cb14\\cb13 \\cf5 a
 \\cb14\\cb13 \\b \\cf5 ,
\\b0
 \\cb14\\cb13 \\cf0  
 \\cb14\\cb13 \\cf5 b
 \\cb14\\cb13 \\b \\cf5 ,
\\b0
 \\cb14\\cb13 \\cf0  
 \\cb14\\cb13 \\cf5 c
 \\cb14\\cb13 \\b \\cf5 )
\\b0
 \\cb14\\cb13 \\b \\cf5 :
\\b0
 \\cb14\\
\\cf5  12 
\\cf2 ... 
\\cf0     
\\b \\cf3 pass
\\b0
\\
}
"""

# =============================================================================

def load_command(mod_name):
    test_file = Path(__file__).resolve()
    scripts = test_file.parent.parent / 'bin'
    scripts = scripts.resolve()
    filename = str(scripts / mod_name)

    # Load the file as if it were a module
    loader = SourceFileLoader(mod_name, filename)
    spec = importlib.util.spec_from_loader(mod_name, loader)
    module = importlib.util.module_from_spec(spec)
    sys.modules[mod_name] = module
    spec.loader.exec_module(module)

    return module

# =============================================================================

class TestCommands(TestCase):
    def test_pat(self):
        display_code = Path(__file__).parent / \
            '../extras/display_code/decorator.repl'
        display_code = str(display_code.resolve())
        sys.argv = ['pat', '--num', '10', '--highlight', '1-2', display_code]

        pat = load_command('pat')

        with capture_stdout() as captured:
            pat.main()

            # Get the output content and remove the ESC characters
            content = captured.getvalue()
            content = content.replace('\x1b', '')

        self.assertEqual(PAT_EXPECTED, content)

    def test_prat(self):
        RTFColourizer.reset_background_colour()

        display_code = Path(__file__).parent / \
            '../extras/display_code/decorator.repl'
        display_code = str(display_code.resolve())
        sys.argv = ['prat', '--num', '10', '--highlight', '1-2', 
            '--background', 'ccaa22', display_code]

        prat = load_command('prat')

        with capture_stdout() as captured:
            prat.main()

            # Get the output content and remove the ESC characters
            content = captured.getvalue()

        self.assertEqual(PRAT_EXPECTED, content)



from pygments.token import Keyword, Name, Comment, String, Error, \
    Number, Operator, Generic, Token, Whitespace, Punctuation

from purdy.parser import FoldedCodeLine, token_ancestor

# =============================================================================
# ANSI Colourizer: augments code with colour specified using ANSI terminal 
#                  escape sequences

_code_palette = {
    #                    fg256         attr
    Token:              ('',           '',),
    Whitespace:         ('',           '',),
    Comment:            ('cyan',       '',),
    Keyword:            ('yellow',     '',),
    Operator:           ('yellow',     '',),
    Name:               ('light_gray', '',),
    Name.Builtin:       ('#006688',    '',),
    Name.Function:      ('light_gray', '',),
    Name.Namespace:     ('light_gray', '',),
    Name.Class:         ('light_gray', '',),
    Name.Exception:     ('dark_green', '',),
    Name.Decorator:     ('#6666dd',    'bold',),
    Name.Variable:      ('',           '',),
    Name.Constant:      ('',           '',),
    Name.Attribute:     ('',           '',),
    Name.Tag:           ('',           '',),
    String:             ('magenta',    '',),
    Number:             ('magenta',    '',),
    Generic.Prompt:     ('dark_blue',  '',),
    Generic.Error:      ('dark_green', '',),
    Generic.Traceback:  ('#aa0000',    'bold',),
    Error:              ('#ff0000',    '',),
}

_xml_palette = dict(_code_palette)
_xml_palette.update({
    Name.Attribute: ('yellow',     '',),
    Keyword:        ('#006688',    '',),
    Name.Tag:       ('#006688',    '',),
    Punctuation:    ('#006688',    '',),
})

import colored

class ANSIColourizer:
    palettes = {
        'code':_code_palette,
        'xml':_xml_palette,
    }

    reset = colored.attr('reset')
    highlight = colored.bg('grey_23')
    get_code = lambda fn, colour: '' if not colour else fn(colour)

    @classmethod
    def colourize(cls, code_line):
        """Returns a text string with ANSI colour escapes corresponding to the
        tokens sent in

        :param code_line: a :class:`CodeLine` object to colourize
        """
        if isinstance(code_line, FoldedCodeLine):
            fg = colored.fg('white')
            bold = colored.attr('bold')
            return f'    {fg}{bold}⋮{cls.reset}'

        palette = code_line.lexer.palette
        ancestor_list = cls.palettes[palette].keys()

        bg = ''
        if code_line.highlight:
            bg = cls.highlight

        output = []
        if code_line.line_number != -1:
            output.append( cls.line_number(code_line.line_number) )

        for part in code_line.parts:
            key = token_ancestor(part.token, ancestor_list)
            colours = cls.palettes[palette][key]

            fg = cls.get_code(colored.fg, colours[0])
            attr = cls.get_code(colored.attr, colours[1])

            output.append(f'{fg}{attr}{bg}{part.text}{cls.reset}')

        return ''.join(output)

    @classmethod
    def line_number(cls, num):
        """Returns a colourized version of a line number"""
        fg = colored.fg('grey_39')
        return f'{fg}{num:3}{cls.reset} '



#!/usr/bin/env python
"""
Animation Steps
===============

An animation is created through a series of steps that are executed together.
A :class:`.cell.GroupCell` wraps these steps. When the user moves forwards and 
backwards through the animations each cell is rendered or undone. This module 
"""
from copy import deepcopy

from pygments.token import String, Token

from purdy.parser import (CodeLine, CodePart, token_is_a, parse_source,
    PurdyLexer)

#import logging
#logging.basicConfig(filename='debug.log', level=logging.DEBUG)
#logger = logging.getLogger()

# ===========================================================================
# Animation Steps
# ===========================================================================

class BaseEditStep:
    def __init__(self, code_box, lines):
        self.code_box = code_box
        self.lines = lines

        if isinstance(lines, CodeLine):
            # if lines is a single item, put it in a list for consistency
            self.lines = [lines, ]

        self.undo_size = len(self.lines)

    @property
    def trunc_lines(self):
        if len(self.lines) == 0:
            return '[]'

        if len(self.lines) == 1:
            return f'[{self.lines[0]}'

        # else len(self.lines) > 1:
        return f'[{self.lines[0]}...]'

# ---------------------------------------------------------------------------
# Line Insertion Steps
# ---------------------------------------------------------------------------

class InsertRows(BaseEditStep):
    def __init__(self, code_box, position, lines):
        super().__init__(code_box, lines)
        self.position = position

    def __str__(self):
        return f'steps.InsertRow("{self.trunc_lines}" @ {self.position})'

    def render_step(self):
        self.code_box.listing.insert_lines(self.position, self.lines)

    def undo_step(self):
        position = self.position
        if position == 0:
            # insert 0 is signal for append, remove last n-lines 
            position = -1 * self.undo_size

        self.code_box.listing.remove_lines(position, self.undo_size)


class Subprocess:
    def __init__(self, code_box, cmd):
        self.code_box = code_box
        self.cmd = cmd

    def __str__(self):
        return f'steps.Subprocess("{self.cmd}")'

    def render_step(self):
        import subprocess

        args = self.cmd.strip().split(' ')
        
        # subprocess 3.7 command is easier to understand, replace code with
        # whats in comment when py3.6 support is dropped
        #
        #result = subprocess.run(args, capture_output=True, text=True)
        result = subprocess.run(args, stdout=subprocess.PIPE,
            stderr=subprocess.PIPE, universal_newlines=True)

        lexer = PurdyLexer.factory_from_name('con')

        new_lines = []
        for row in result.stdout.split('\n'):
            part = CodePart(Token.Generic.Output, row)
            new_lines.append( CodeLine([part, ], lexer) )

        self.undo_position = len(self.code_box.listing.lines) + 1
        self.undo_size = len(new_lines)
        self.code_box.listing.insert_lines(0, new_lines)

    def undo_step(self):
        self.code_box.listing.remove_lines(self.undo_position, self.undo_size)

# ---------------------------------------------------------------------------
# Line Editing Steps
# ---------------------------------------------------------------------------

class ReplaceRows(BaseEditStep):
    def __init__(self, code_box, position, lines):
        super().__init__(code_box, lines)
        self.position = position

    def __str__(self):
        return f'steps.ReplaceRows("{self.trunc_lines}" @ {self.position})'

    def render_step(self):
        real_position = self.code_box.listing.positive_position(self.position)
        box_length = len(self.code_box.listing.lines)
        size = len(self.lines)
        if real_position + size > box_length + 1:
            raise AttributeError( ('Replacement happens in place. You cannot '
                'replace outside the code_box. You are attempting to '
                f'overwrite {size} lines at position {real_position} in a '
                f'code_box that only has {box_length} lines') )

        self.undo_lines = self.code_box.listing.copy_lines(self.position,
            self.undo_size)

        for count, line in enumerate(self.lines):
            self.code_box.listing.replace_line(self.position + count, line)

    def undo_step(self):
        for count, line in enumerate(self.undo_lines):
            self.code_box.listing.replace_line(self.position + count, line)


class SuffixRow(BaseEditStep):
    def __init__(self, code_box, position, source, cursor=False):
        self.code_box = code_box
        self.position = position
        self.source = source
        self.cursor = cursor

    def __str__(self):
        return f'steps.SuffixRow("{self.source}" @ {self.position})'

    def _inside_string(self, line):
        # Returns True if appending something to the end of this line will be
        # appending inside of a string
        last_token = line.parts[-1][0]
        if not token_is_a(last_token, String):
            return False

        # token is a String, check if multi-line
        try:
            # (position is 1-indexed, so next line in 0-index is itself)
            next_line = self.code_box.listing.lines[self.position]
        except IndexError:
            # nothing after this, so it must be a closed string
            return False

        last_token_next_line = next_line.parts[0][0]
        if not token_is_a(last_token_next_line, String):
            # next token is not a string, must be a closed string
            return False

        # if you get here we are a String and we have a String as the next
        # line, this could be because we're inside a multi-line string or
        # because the code has two doc strings back-to-back, this is the worst
        # case scenario, loop through the code and count String openers that
        # are the same as this one, if odd number, we're inside a string
        inside = False
        for line in self.code_box.listing.lines[0:self.position]:
            for part in line.parts:
                if part.token == last_token:
                    inside = not inside

        return inside

    def render_step(self):
        line = deepcopy(self.code_box.listing.lines[self.position - 1])
        self.undo_line = line

        token = line.parts[-1][0]
        if self._inside_string(line):
            # inside a multi-line string, don't reparse, just append
            parts = deepcopy(line.parts)
            parts[-1] = CodePart(token, parts[-1].text + self.source)
            replace_line = CodeLine(parts, line.lexer)
        else:
            # not inside a string
            if token_is_a(token, String):
                # last token is a string, parse the source and append
                source_line = parse_source(self.source, line.lexer)[0]
                new_parts = line.parts + source_line.parts
                replace_line = CodeLine(new_parts, line.lexer)
            else:
                # last token isn't a string, reparse the whole line
                text = line.text + self.source
                replace_line = parse_source(text, line.lexer)[0]

        if self.cursor:
            replace_line.parts.append( CodePart(Token, '\u2588') )

        self.code_box.listing.replace_line(self.position, replace_line)

    def undo_step(self):
        self.code_box.listing.replace_line(self.position, self.undo_line)


class RemoveRows:
    def __init__(self, code_box, position, num=1):
        self.code_box = code_box
        self.position = position
        self.num = num

    def __str__(self):
        return f'steps.RemoveRows({self.position} for {self.num})'

    def render_step(self):
        self.undo_lines = []
        for x in range(0, self.num):
            line = self.code_box.listing.lines[x + self.position - 1]
            undo_line = deepcopy(line)
            self.undo_lines.append(undo_line)

        self.code_box.listing.remove_lines(self.position, self.num)

    def undo_step(self):
        self.code_box.listing.insert_lines(self.position, self.undo_lines)


class Clear:
    def __init__(self, code_box):
        self.code_box = code_box

    def __str__(self):
        return 'steps.Clear()'

    def render_step(self):
        self.undo_lines = []
        for line in self.code_box.listing.lines:
            undo_line = deepcopy(line)
            self.undo_lines.append(undo_line)

        self.code_box.listing.clear()

    def undo_step(self):
        self.code_box.listing.insert_lines(0, self.undo_lines)

# ---------------------------------------------------------------------------
# Presentation Steps
# ---------------------------------------------------------------------------

class HighlightLines:
    def __init__(self, code_box, numbers, highlight_on):
        self.code_box = code_box
        if isinstance(numbers, int):
            self.numbers = [numbers, ]
        else:
            self.numbers = numbers
        self.highlight_on = highlight_on

    def __str__(self):
        return f'steps.HighlightLines("{self.numbers}", {self.highlight_on})'

    def _set_highlight(self, highlight_value):
        for num in self.numbers:
            self.code_box.listing.set_highlight(highlight_value, num, num)

    def render_step(self):
        self._set_highlight(self.highlight_on)

    def undo_step(self):
        self._set_highlight(not self.highlight_on)


class FoldLines:
    def __init__(self, code_box, position, end):
        self.code_box = code_box
        self.position = position
        self.end = end

    def __str__(self):
        return f'steps.FoldLines({self.position}, {self.end})'

    def render_step(self):
        self.undo_lines = []

        start = self.position - 1
        last = self.end

        if self.end == -1:
            last = len(self.code_box.listing.lines)

        for line in self.code_box.listing.lines[start:last]:
            undo_line = deepcopy(line)
            self.undo_lines.append(undo_line)

        self.code_box.listing.fold_lines(self.position, last)

    def undo_step(self):
        self.code_box.listing.insert_lines(self.position, self.undo_lines)
        self.code_box.listing.remove_lines(self.position + len(self.undo_lines),
            1)

# ---------------------------------------------------------------------------
# Control Steps
# ---------------------------------------------------------------------------

class Sleep:
    def __init__(self, time):
        self.time = time

    def __str__(self):
        return 'steps.Sleep()'


class CellEnd:
    pass


class SectionBreak:
    pass


class Transition:
    def __init__(self, code_box, code=None, code_box_to_copy=None):
        self.code_box = code_box
        self.code = code
        self.code_box_to_copy = code_box_to_copy

    def __str__(self):
        # Import here to avoid circular import
        from purdy.actions import condense
        code_content = ''
        if self.code:
            code_content = condense(self.code.source)

        copy_content = ''
        if self.code_box_to_copy:
            lines = [str(line) for line in self.code_box_to_copy.listing.lines]
            copy_content = condense('\n'.join(lines))

        return f'steps.Transition("{code_content}", "{copy_content}")'


class StopMovieException(Exception):
    pass


class StopMovie:
    def __str__(self):
        return 'steps.StopMovie()'

    def render_step(self):
        raise StopMovieException()

    def undo_step(self):
        pass


class RunFunction:
    def __init__(self, fn, undo, *args, **kwargs):
        self.fn = fn
        self.undo = undo
        self.fn_args = args
        self.fn_kwargs = kwargs

    def __str__(self):
        undo_name = 'None'
        if self.undo:
            undo_name = self.undo.__name__

        return (f'steps.RunFunction({self.fn.__name__}, {undo_name},'
            f' args={self.fn_args}, kwargs={self.fn_kwargs})')

    def render_step(self):
        self.fn(*self.fn_args, **self.fn_kwargs)

    def undo_step(self):
        if self.undo is not None:
            self.undo(*self.fn_args, **self.fn_kwargs)



#!/usr/bin/env python
"""
Animation Management
====================

This module handles the slide rendering animation in the urwid purdy player
"""
from enum import Enum

#import logging
#logging.basicConfig(filename='debug.log', level=logging.DEBUG)
#logger = logging.getLogger()

# ==========================================================================
# Animation Manager
# ==========================================================================

class AnimationManager:
    class State(Enum):
        ACTIVE = 0
        SLEEPING = 1

    handled_keys = ['left', 'right', 's', 'S', '0', '1', '2', '3', '4', '5',
        '6', '7', '8', '9', 'esc']

    def __init__(self, screen):
        self.screen = screen
        self.cells = []
        self.index = -1
        self.multi_action = ''

    def register(self, content):
        """Register one or more "cell" classes from 
        :mod:`purdy.animation.cell` objects to be animated. 

        :param content: a single Cell or iterable of them
        """
        if isinstance(content, list) or isinstance(content, tuple):
            for cell in content:
                self.cells.append(cell)
        else:
            self.cells.append(content)

    def _conditional_alarms(self, cell):
        # if we're in movie mode and the cell isn't doing a typewriter
        # animation, wake us up for the next animation
        if not cell.is_animating and self.screen.movie_mode > -1:
            self.screen.set_alarm('movie_alarm', self.screen.movie_mode)

    def auto_forward_alarm(self):
        self.state = self.State.ACTIVE
        self.forward()

    def animation_alarm(self):
        self.state = self.State.ACTIVE
        cell = self.cells[self.index]
        cell.animation_wake_up(self)
        self._conditional_alarms(cell)

    def movie_alarm(self):
        self.State.ACTIVE
        self.forward()
        
    def perform(self, key):
        if key == 's' and self.state == self.State.SLEEPING:
            self.interrupt()
            return

        if self.state == self.State.SLEEPING:
            # ignore key presses while asleep
            return

        if key == 'right':
            self.forward()
        elif key == 'left':
            self.backward()
        elif key == 's':
            self.fast_forward()
        elif key == 'S':
            self.next_section()
        elif key in ('0', '1', '2', '3', '4', '5', '6', '7', '8', '9'):
            self.multi_action += key
        elif key == 'esc':
            self.multi_action = ''

    def forward(self):
        if self.index + 1 >= len(self.cells):
            return

        self.index += 1
        cell = self.cells[self.index]
        cell.render(self)

        if not cell.auto_forward:
            self._conditional_alarms(cell)

    def fast_forward(self):
        repeat = 1
        if self.multi_action:
            repeat = int(self.multi_action)
            self.multi_action = ''

        while self.index + 1 < len(self.cells) and repeat > 0:
            self.index += 1
            self.cells[self.index].render(self, skip=True)
            repeat -= 1

    def next_section(self):
        repeat = 1
        if self.multi_action:
            repeat = int(self.multi_action)
            self.multi_action = ''

        while self.index + 1 < len(self.cells) and repeat > 0:
            self.index += 1
            self.cells[self.index].render(self, skip=True)

            if self.cells[self.index].section_break:
                # found the boundary, stop
                repeat -= 1

    def interrupt(self):
        self.cells[self.index].interrupt(self)

    def backward(self):
        repeat = 1
        if self.multi_action:
            repeat = int(self.multi_action)
            self.multi_action = ''

        while self.index > 0 and repeat > 0:
            self.cells[self.index].undo(self)
            self.index -= 1
            repeat -= 1



#!/usr/bin/env python
"""
Animation Cells
===============

A Cell represents an animation used by the
:class:`purdy.animation.manager.AnimationManager`. A Cell is responsible for
rendering or undoing work to a :class:`purdy.widget.CodeBox`. Animating cells
can partially render then wait for an alarm provided by the manager to continue
rendering.
"""
from copy import deepcopy
from enum import Enum
import random

from purdy.animation import steps as steplib
from purdy.parser import BlankCodeLine, parse_source

# ===========================================================================
# Cell Factory
# ===========================================================================

def group_steps_into_cells(steps):
    """Actions create multiple steps possibly with cell breaks between them.
    This method takes a list of steps and returns a list of Cell objects,
    grouping the steps by break marker

    :param steps: a list of steps
    :returns: a list of Cell objects
    """
    cells = []
    cell = None
    previous_step = None
    for step in steps:
        if isinstance(step, steplib.CellEnd):
            if cell:
                cells.append(cell)

            if isinstance(previous_step, steplib.Transition):
                # By default Transition automatically goes to next step, a
                # wait is ignored because Transition itself is a cell, in this
                # case need to change the behaviour of Transition
                cells[-1].auto_forward = False

            cell = None
        elif isinstance(step, steplib.Transition):
            if cell:
                cells.append(cell)

            cell = TransitionCell(step.code_box, step.code,
                step.code_box_to_copy)

            cells.append(cell)
            cell = None
        elif isinstance(step, steplib.SectionBreak):
            # Section markers are't cells, they just change the state of the
            # previous cell
            if cells:
                cells[-1].section_break = True
        else:
            if not cell:
                cell = GroupCell()

            cell.steps.append(step)

        previous_step = step

    if cell and len(cell.steps) > 0:
        cells.append(cell)

    return cells

# ===========================================================================
# Base Classes
# ===========================================================================

class AnimatingCellBase:
    @property
    def is_animating(self):
        return self.animation_alarm_handle is not None

    def animation_wake_up(self, manager):
        self.animation_alarm_handle = None
        self.render(manager)

    def interrupt(self, manager):
        # interrupted during an animation sequence, remove the timer
        manager.screen.remove_alarm(self.animation_alarm_handle)

        self.animation_alarm_handle = None
        manager.state = manager.State.ACTIVE
        self.render(manager, skip=True)

# ===========================================================================
# Cells
# ===========================================================================

class GroupCell(AnimatingCellBase):
    """Groups steps together into a bundle that can be rendered or undone
    together. Implements animation alarms so the manager can do timed call
    backs into the group and continue rendering.
    """
    auto_forward = False

    def __init__(self):
        self.steps = []
        self.index = 0
        self.animation_alarm_handle = None
        self.section_break = False

    def __str__(self):
        step_string = ','.join([str(step) for step in self.steps])
        return 'GroupCell(steps=[' + step_string + '])'

    def render(self, manager, skip=False):
        if len(self.steps) == 0 or self.index >= len(self.steps):
            # badly formed action sequences that result in an empty cell would
            # cause a crash, just do nothing
            return

        for x in range(self.index, len(self.steps)):
            self.index = x
            step = self.steps[x]

            if isinstance(step, steplib.Sleep):
                if skip:
                    # in fast-forward mode, do nothing with this Sleep
                    continue
                else:
                    manager.state = manager.State.SLEEPING

                    time = step.time
                    if isinstance(step.time, tuple):
                        time = random.uniform(step.time[0], step.time[1])

                    self.animation_alarm_handle = manager.screen.set_alarm(
                        'animation_alarm', time)

                    self.index += 1
                    return

            try:
                step.render_step()
            except steplib.StopMovieException:
                manager.screen.movie_mode = -1

    def undo(self, manager):
        if len(self.steps) == 0:
            # badly formed action sequences that result in an empty cell would
            # cause a crash, just do nothing
            return

        # if we rendered everything and the last call was a Sleep, the index
        # may be past our list bounds
        if self.index >= len(self.steps):
            self.index = len(self.steps) - 1

        for x in range(self.index, -1, -1):
            self.index = x
            step = self.steps[x]
            if not isinstance(step, steplib.Sleep):
                step.undo_step()


class TransitionCell(AnimatingCellBase):
    DELAY = 0.05
    BETWEEN_DELAY = 0.3

    class State(Enum):
        BEFORE = 0
        DELETING = 1
        APPENDING = 2
        DONE = 3

    def __init__(self, code_box, code, code_box_to_copy):
        self.code_box = code_box
        self.code = code
        self.code_box_to_copy = code_box_to_copy
        self.state = self.State.BEFORE
        self.steps = []
        self.auto_forward = True
        self.section_break = False

    def render(self, manager, skip=False):
        if self.state == self.State.DONE:
            return

        manager.state = manager.State.SLEEPING
        delay = self.DELAY
        if self.state == self.State.BEFORE:
            # first time through, setup our append and undo lines
            self.position = 1

            if self.code:
                self.code_lines = parse_source(self.code.source, 
                    self.code.lexer)
            elif self.code_box_to_copy:
                self.code_lines = deepcopy(self.code_box_to_copy.listing.lines)
            else:
                self.code_lines = []

            self.undo_lines = deepcopy(self.code_box.listing.lines)
            self.state = self.State.DELETING
            # intentional fall through

        # If we're in skip mode, need to keep doing the work
        while True:
            if self.state == self.State.DELETING:
                if self.position >= len(self.code_box.listing.lines):
                    # wiped everything, delete the empty boxes and go into
                    # append mode
                    self.code_box.listing.clear()
                    self.state = self.State.APPENDING
                    delay = self.BETWEEN_DELAY
                else:
                    line = BlankCodeLine()
                    self.code_box.listing.replace_line(self.position, line)
                    self.position += 1
            else: # state == APPENDING
                try:
                    line = self.code_lines.pop(0)
                    self.code_box.listing.insert_lines(0, [line, ])
                except IndexError:
                    manager.state = manager.State.ACTIVE
                    self.state = self.State.DONE
                    if self.auto_forward:
                        manager.screen.set_alarm('auto_forward_alarm', 0)
                    return

            if not skip:
                # Not skipping, do the loop only once and get out
                break

        # ask to be woken back up for the next step
        self.animation_alarm_handle = manager.screen.set_alarm(
            'animation_alarm', delay)

    def undo(self, manager):
        self.state = self.State.BEFORE
        self.code_box.listing.clear()
        self.code_box.listing.insert_lines(0, self.undo_lines)



from copy import deepcopy
from unittest.mock import patch, mock_open, Mock, call

from pygments.token import Token

from purdy.content import Code, Listing
from purdy.parser import CodeLine, CodePart, FoldedCodeLine

from tests.base import (PurdyContentTest, py3_lexer, PY_CODE_LINES,
    BASH_CODE_LINES)

# =============================================================================

def py_line(text, num=-1):
    line = CodeLine([CodePart(Token.Text, text)], py3_lexer)
    if num > 0:
        line.line_number = num
    return line

def alpha_lines(num):
    # returns "num" code lines, with contents A, B, C...
    return [
        CodeLine([CodePart(Token.Text, chr(x))], py3_lexer) for x in \
            range(65, 65 + num)
    ]

# commented out to hide from pyflakes
#
#def _print_lines(title, lines):
#    # Used to help debug the tests
#    print(f'==== {title} ====')
#    for line in lines:
#        print(line)


class TestContent(PurdyContentTest):
    def assert_code_parts(self, line_list, listing):
        # loop through line_list (consists of manually created CodeLine
        # objects) and compare it to the actually parsed code object
        for i, line in enumerate(listing.lines):
            for j, part in enumerate(line.parts):
                expected = line_list[i].parts[j]

                self.assertEqual(expected.token, part.token,
                    msg=f'Failed for CodeLine[{i}]')
                self.assertEqual(expected.text, part.text,
                    msg=f'Failed for CodeLine[{i}]')

    def assert_line_text(self, expected_text, lines):
        self.assertEqual(len(expected_text), len(lines), 
            msg='Size mismatch between expected_text and lines')
        for x, text in enumerate(expected_text):
            self.assertEqual(text, lines[x].parts[0].text,
                msg=f'expected_text[{x}] did not match line')

    def hooked_listing(self, append_code=True):
        mock_hook = Mock()
        listing = Listing(starting_line_number=10)
        listing.set_display('ansi', mock_hook)

        if append_code:
            code = Code(text=self.py_source, lexer_name='py3')
            listing.append_code(code)

        return mock_hook, listing

    #=== Test Methods

    def test_construction(self):
        #---Test consttruction via a Code object. Need to test both py3 and
        # bash to get all parsing cases
        with patch('builtins.open', mock_open(read_data=self.bash_source)):
            code = Code(filename='foo', lexer_name='bash')
            listing = Listing(code)
            self.assert_code_parts(BASH_CODE_LINES, listing)

        code = Code(text=self.py_source)  # py3 lexer, but let it auto-detect
        listing = Listing(code)
        self.assert_code_parts(PY_CODE_LINES, listing)

        #---Test line numbering
        listing.starting_line_number = 10
        listing.reset_line_numbers()
        for count, line in enumerate(listing.lines):
            self.assertEqual(count + 10, line.line_number)

        #---Test lists of Code 
        listing = Listing([code, code])
        self.assertEqual(2 * len(PY_CODE_LINES), len(listing.lines))

    def test_setters(self):
        mock_hook, listing = self.hooked_listing()

        new_lines = [py_line('foo'), py_line('bar')]

        #--- Create a new listing, append code and check render hook was called
        calls = [ call(listing, i+1, line) for i, line in \
            enumerate(listing.lines)]
        mock_hook.line_inserted.assert_has_calls(calls)

        # Test mid insertion 
        size = len(listing.lines)
        mock_hook.reset_mock()
        listing.insert_lines(2, new_lines)

        self.assertEqual(py_line('foo', 11), listing.lines[1])
        self.assertEqual(py_line('bar', 12), listing.lines[2])
        self.assertEqual(size + 2, len(listing.lines))

        calls = [ 
            call(listing, 2, listing.lines[1]), 
            call(listing, 3, listing.lines[2])
        ]
        mock_hook.line_inserted.assert_has_calls(calls)

        # ensure line numbering got reset
        for count, line in enumerate(listing.lines):
            self.assertEqual(count + 10, line.line_number)

        # ensure the change call was made for the lines whose numbers changed
        calls = []
        for x in range(3, len(listing.lines)):
            calls.append( call(listing, x + 1, listing.lines[x]) )
        mock_hook.line_changed.assert_has_calls(calls)

        #--- Test negative position insertion
        size = len(listing.lines)
        mock_hook.reset_mock()

        listing.insert_lines(-1, new_lines)

        calls = [ 
            call(listing, size, listing.lines[-3]), 
            call(listing, size + 1, listing.lines[-2])
        ]
        mock_hook.line_inserted.assert_has_calls(calls)

        self.assertEqual(size + 2, len(listing.lines))
        self.assertEqual(py_line('foo', 18), listing.lines[-3])
        self.assertEqual(py_line('bar', 19), listing.lines[-2])

        #--- Test replacement
        size = len(listing.lines)
        mock_hook.reset_mock()
        replace_line = py_line('waz')

        listing.replace_line(-1, replace_line)

        mock_hook.line_changed.assert_called_once_with(listing, size,
            listing.lines[-1])

        self.assertEqual(size, len(listing.lines))
        self.assertEqual(py_line('waz', 20), listing.lines[-1])

    def test_remove(self):
        mock_hook, listing = self.hooked_listing()

        expected = deepcopy(listing.lines)
        del expected[1]
        del expected[1]
        for count, line in enumerate(expected):
            line.line_number = 10 + count

        listing.remove_lines(2, 2)
        self.assertEqual(expected, listing.lines)

        # Verify removal hooks
        calls = [ 
            call(listing, 2),
            call(listing, 2),
        ]
        mock_hook.line_removed.assert_has_calls(calls)

        # Check lines were renumbered
        listing.starting_line_number = 10
        for count, line in enumerate(listing.lines):
            self.assertEqual(count + 10, line.line_number)

        # ensure the change call was made for the lines whose numbers changed
        calls = []
        for x in range(1, len(listing.lines)):
            calls.append( call(listing, x + 1, listing.lines[x]) )
        mock_hook.line_changed.assert_has_calls(calls)

    def test_clear(self):
        mock_hook, listing = self.hooked_listing()

        listing.clear()
        self.assertEqual(0, len(listing.lines))
        mock_hook.clear.assert_called_once()

    def test_highlight(self):
        mock_hook, listing = self.hooked_listing()

        listing.set_highlight(True, 2, 3)
        for count, line in enumerate(listing.lines):
            if count in (1, 2):
                self.assertTrue(line.highlight)
            else:
                self.assertFalse(line.highlight)

        listing.set_highlight(False, 1)
        for count, line in enumerate(listing.lines):
            self.assertFalse(line.highlight)

    def test_content(self):
        code = Code(text=self.py_source, lexer_name='py3')
        listing = Listing(code, starting_line_number=10)

        result = listing.content()
        for count, line in enumerate(listing.lines):
            self.assertEqual(result[count], f'{count+10:3} ' + line.text)

    def test_insert_lines(self):
        ### Test insert with various positions

        # Empty and 0
        listing = Listing(starting_line_number=10)
        listing.insert_lines(0, alpha_lines(1))
        self.assertEqual(1, len(listing.lines))

        # One line and 0
        listing.insert_lines(0, [py_line('B'), ])
        self.assertEqual(2, len(listing.lines))
        self.assert_line_text('AB', listing.lines)

        # Three lines and -1
        listing = Listing(starting_line_number=10)
        listing.insert_lines(0, alpha_lines(3))

        listing.insert_lines(-1, [py_line('Z'), ])
        self.assert_line_text('ABZC', listing.lines)

        # Inserting multiple midway
        listing = Listing(starting_line_number=10)
        listing.insert_lines(0, alpha_lines(3))

        listing.insert_lines(2, [ py_line('X'), py_line('Y'), py_line('Z'), ])
        self.assert_line_text('AXYZBC', listing.lines)

        # Boundary tests
        listing = Listing(starting_line_number=10)
        listing.insert_lines(0, alpha_lines(1))

        with self.assertRaises(IndexError):
            listing.insert_lines(-10, [py_line('Z'), ])

        with self.assertRaises(IndexError):
            listing.insert_lines(100, [py_line('Z'), ])

    def test_replace_line(self):
        ### Test replace with various positions

        # One line and 1
        listing = Listing(starting_line_number=10)
        listing.insert_lines(0, alpha_lines(1))
        listing.replace_line(1, py_line('Z'))
        self.assert_line_text('Z', listing.lines)

        # One line and -1
        listing = Listing(starting_line_number=10)
        listing.insert_lines(0, alpha_lines(1))
        listing.replace_line(-1, py_line('Z'))
        self.assert_line_text('Z', listing.lines)

        # Three lines and 2 
        listing = Listing(starting_line_number=10)
        listing.insert_lines(0, alpha_lines(3))
        listing.replace_line(2, py_line('Z'))
        self.assert_line_text('AZC', listing.lines)

        # Three lines and -2
        listing = Listing(starting_line_number=10)
        listing.insert_lines(0, alpha_lines(3))
        listing.replace_line(-2, py_line('Z'))
        self.assert_line_text('AZC', listing.lines)

        # Boundary test
        listing = Listing(starting_line_number=10)
        listing.insert_lines(0, alpha_lines(1))

        with self.assertRaises(IndexError):
            listing.replace_line(0, py_line('Z'))

        with self.assertRaises(IndexError):
            listing.replace_line(-10, py_line('Z'))

        with self.assertRaises(IndexError):
            listing.replace_line(100, py_line('Z'))

    def test_remove_lines(self):
        ### Test remove with various positions

        # Two lines, remove @ 1
        listing = Listing(starting_line_number=10)
        listing.insert_lines(0, alpha_lines(2))
        listing.remove_lines(1)
        self.assert_line_text('B', listing.lines)

        # Three lines, remove @ 2 
        listing = Listing(starting_line_number=10)
        listing.insert_lines(0, alpha_lines(3))
        listing.remove_lines(2)
        self.assert_line_text('AC', listing.lines)

        # Three lines and -2
        listing = Listing(starting_line_number=10)
        listing.insert_lines(0, alpha_lines(3))
        listing.remove_lines(-2)
        self.assert_line_text('AC', listing.lines)

        #---Remove multiple lines

        # from top
        listing = Listing(starting_line_number=10)
        listing.insert_lines(0, alpha_lines(4))
        listing.remove_lines(1, 2)
        self.assert_line_text('CD', listing.lines)

        # from middle
        listing = Listing(starting_line_number=10)
        listing.insert_lines(0, alpha_lines(4))
        listing.remove_lines(-3, 2)
        self.assert_line_text('AD', listing.lines)

        #---Boundary test
        listing = Listing(starting_line_number=10)
        listing.insert_lines(0, alpha_lines(1))

        with self.assertRaises(IndexError):
            listing.remove_lines(0)

        with self.assertRaises(IndexError):
            listing.remove_lines(-10)

        with self.assertRaises(IndexError):
            listing.remove_lines(100)

    def test_copy_lines(self):
        listing = Listing(starting_line_number=10)
        listing.insert_lines(0, alpha_lines(4))

        result = listing.copy_lines(2, 2)
        self.assert_line_text('BC', result)

    def test_fold_lines(self):
        mock_hook, listing = self.hooked_listing()

        #---Typical usage, folding some lines
        listing.fold_lines(4, 6)

        self.assertEqual(5, len(listing.lines))
        self.assertEqual(10, listing.lines[0].line_number)
        self.assertEqual(11, listing.lines[1].line_number)
        self.assertEqual(12, listing.lines[2].line_number)
        self.assertEqual(16, listing.lines[4].line_number)

        self.assertTrue( isinstance(listing.lines[3], FoldedCodeLine) )

        # Verify hooks
        calls = [ 
            call(listing, 5),
            call(listing, 5),
        ]
        mock_hook.line_removed.assert_has_calls(calls)
        mock_hook.line_changed.assert_called_once()

        #---Fold testing different positions and parms

        # First line
        listing = Listing(starting_line_number=10)
        listing.insert_lines(0, alpha_lines(4))

        listing.fold_lines(1, 2)
        self.assertEqual( FoldedCodeLine(2), listing.lines[0])
        self.assert_line_text('CD', listing.lines[1:])

        # One line
        listing = Listing(starting_line_number=10)
        listing.insert_lines(0, alpha_lines(4))

        listing.fold_lines(1, 1)
        self.assertEqual( FoldedCodeLine(1), listing.lines[0])
        self.assert_line_text('BCD', listing.lines[1:])



"""
Settings (purdy.settings.py)
============================

Defines the default settings for :class:`purdy.ui.Screen` objects, can be
overridden by passing an altered dictionary into the Screen's constructor. 
"""

settings = {
    # delay between characters appearing on screen (in milliseconds)
    'delay':130,

    # range of random time in milliseconds to change the delay; makes typing
    # look more natural
    'delay_variance':30,

    # movie mode: instead of waiting for key presses, play like a movie, -1
    # disables, otherwise value in milliseconds for delay between played steps
    # (like 'delay' field
    'movie_mode':-1,

    # xterm colour mode, anything but 256 gives 16 colour mode
    'colour':256,

    # if True, stops Screen from running argparse
    'deactivate_args':False,

    # max height for presentation, only works in TUI mode, 0 == no max
    'max_height':0,
}



"""
Actions
=======

Library users specify a series of actions that turn into the presentation
animations in the Urwid client.  An action is similar to a slide in a slide
show, except it can both present and change lines of code on the screen.

All purdy library programs have the following basic structure:

.. code-block:: python

    screen = Screen(...)
    actions = [ ... ]
    screen.run(actions)

Each action gets translated into a series of steps defined in the
:mod:`purdy.animation` module.
"""
import random
from copy import deepcopy

from pygments.token import Generic, Token

from purdy.animation import steps as steplib
from purdy.parser import CodePart, CodeLine, parse_source, token_is_a
from purdy.scribe import range_set_to_list

# =============================================================================

def condense(value):
    if not value:
        return ''

    content = deepcopy(value)
    if len(content) >= 20:
        content = content[0:17] + '...'

    return content

# =============================================================================
# Single Code Blob Actions
# =============================================================================

class Insert:
    """Inserts the content of a :class:`purdy.content.Code` object to a
    specified line in a :class:`purdy.ui.CodeBox`. Pushes content down,
    inserting at "1" is the beginning of the list. Position is 1-indexed

    :param code_box: the :class:`purdy.ui.CodeBox` instance to insert code
                     into

    :param position: line number to insert code at. Position is 1-indexed.
                     Content is pushed down, so a value of "1" inserts at the
                     beginning. Negative indicies are supported. A value of
                     "0" will append the code to the bottom.

    :param code: a :class:`purdy.content.Code` object containing the source
                 code to insert.
    """
    def __init__(self, code_box, position, code):
        self.code_box = code_box
        self.code = code
        self.position = position

    def __str__(self):
        content = condense(self.code.source)
        return f'actions.Insert({self.position}, "{content}")'

    def steps(self):
        lines = parse_source(self.code.source, self.code.lexer)
        steps = [steplib.InsertRows(self.code_box, self.position, lines), ]

        return steps


class Append(Insert):
    """Adds the content of a :class:`purdy.content.Code` object to the end of
    a :class:`purdy.ui.CodeBox`.

    :param code_box: the :class:`purdy.ui.CodeBox` instance to insert code
                     into

    :param code: a :class:`purdy.content.Code` object containing the source
                 code to insert.
    """
    def __init__(self, code_box, code):
        super().__init__(code_box, 0, code)

    def __str__(self):
        content = condense(self.code.source)
        return f'actions.Append("{content}")'


class Replace:
    """Replaces one or more lines of a :class:`purdy.ui.CodeBox` using the
    content of a :class:`purdy.content.Code` object. This action attempts to
    overwrite using the number of lines in the :class:`purdy.content.Code`
    object passed in, it is up to you to make sure there is enough space in
    your CodeBox.

    :param code_box: the :class:`purdy.ui.CodeBox` instance where the code is
                     to be replaced

    :param position: line number to replace the code at. Position is 1-indexed.
                     Negative indicies are supported.

    :param code: a :class:`purdy.content.Code` object containing the source
                 code to insert.
    """
    def __init__(self, code_box, position, code):
        self.code_box = code_box
        self.code = code
        self.position = position

    def __str__(self):
        content = condense(self.code.source)
        return f'actions.Replace({self.position}, "{content}")'

    def steps(self):
        lines = parse_source(self.code.source, self.code.lexer)
        steps = [steplib.ReplaceRows(self.code_box, self.position, lines), ]

        return steps


class Remove:
    """Removes one or more lines of a :class:`purdy.ui.CodeBox`.

    :param code_box: the :class:`purdy.ui.CodeBox` instance where the code is
                     to be replaced

    :param position: line number to replace the code at. Position is 1-indexed.
                     Negative indicies are supported.

    :param size: number of lines to remove.
    """
    def __init__(self, code_box, position, size):
        self.code_box = code_box
        self.position = position
        self.size = size

    def __str__(self):
        return f'actions.Remove({self.position}, "{self.size}")'

    def steps(self):
        return [steplib.RemoveRows(self.code_box, self.position, self.size), ]


class Clear:
    """Clears the contents of a :class:`purdy.ui.CodeBox`.

    :param code_box: the :class:`purdy.ui.CodeBox` instance where the code is
                     to be replaced
    """
    def __init__(self, code_box):
        self.code_box = code_box

    def __str__(self):
        return 'actions.Clear()'

    def steps(self):
        return [steplib.Clear(self.code_box), ]

# =============================================================================
# Source Based Actions
# =============================================================================

class Suffix:
    """Adds the provided text to the end of an existing line in a
    :class:`purdy.ui.CodeBox`.

    :param code_box: the :class:`purdy.ui.CodeBox` instance where the code is
                     to be appended

    :param position: line number to replace the code at. Position is 1-indexed.
                     Negative indicies are supported.

    :param source: string containing content to append to the line
    """
    def __init__(self, code_box, position, source):
        self.code_box = code_box
        self.position = position
        self.source = source

    def __str__(self):
        return f'actions.Suffix({self.position}, "{self.source}")'

    def steps(self):
        return [steplib.SuffixRow(self.code_box, self.position, self.source), ]


class Shell:
    """Runs a shell command via subprocess. Does not display the command
    (you're better off using a typewriter command to show it, then use this to
    spit out the results).  Command and results are added to the
    :class:`purdy.ui.CodeBox`.

    :param code_box: the :class:`purdy.ui.CodeBox` instance where the code is
                     to be appended

    :param cmd: string containing the shell command and its paramters.
                Example: ``ls -la``.
    """
    def __init__(self, code_box, cmd):
        self.code_box = code_box
        self.cmd = cmd

    def __str__(self):
        return f'actions.Suffix("{self.cmd}")'

    def steps(self):
        return [steplib.Subprocess(self.code_box, self.cmd), ]

# =============================================================================
# Typewriter Actions
# =============================================================================

class TypewriterBase:
    continuous = [Generic.Prompt, Generic.Output, Generic.Traceback]

    @property
    def delay_until_next_letter(self):
        typing_delay = self.code_box.screen.settings['delay'] / 1000
        variance = self.code_box.screen.settings['delay_variance']

        vary_by = random.randint(0, 2 * variance) - variance
        return typing_delay + (vary_by / 1000)


class TypewriterStep(TypewriterBase):
    def _line_to_steps(self, line, insert_pos, replace_pos):
        steps = []

        # --- Skip animation for "output" content
        first_token = line.parts[0].token
        is_console = self.code.lexer.is_console

        if is_console and not token_is_a(first_token, Generic.Prompt):
            # in console mode only lines with prompts get typewriter
            # animation, everything else is just added directly
            return [steplib.InsertRows(self.code_box, insert_pos, line), ]

        # --- Typewriter animation
        # insert a blank row first with contents of line changing what is on
        # it as animation continues
        dummy_parts = [CodePart(Token, ''), ]
        row_line = CodeLine(dummy_parts, self.code.lexer)
        step = steplib.InsertRows(self.code_box, insert_pos, row_line)
        steps.append(step)

        current_parts = []
        num_parts = len(line.parts)
        for count, part in enumerate(line.parts):
            if part.token in self.continuous:
                # part is a chunk that gets output all together, replace the
                # dummy line with the whole contents
                current_parts.append(part)
                row_line = CodeLine(deepcopy(current_parts), self.code.lexer)
                step = steplib.ReplaceRows(self.code_box, replace_pos, row_line)
                steps.append(step)

                if part.token == Generic.Prompt:
                    # stop animation if this is a prompt, wait for keypress
                    steps.append( steplib.CellEnd() )
            elif count == 0 and token_is_a(part.token, Token.Text) and (
                    part.text.rstrip() == ''):
                # first token is leading whitespace, don't animate it, just
                # insert it
                current_parts.append(part)
                row_line = CodeLine(deepcopy(current_parts), self.code.lexer)
                step = steplib.ReplaceRows(self.code_box, replace_pos, row_line)
                steps.append(step)
            else:
                new_part = CodePart(part.token, '')
                current_parts.append(new_part)

                typewriter = ''
                for letter in part.text:
                    typewriter += letter
                    new_part = CodePart(part.token, typewriter)
                    current_parts[-1] = new_part
                    output_parts = deepcopy(current_parts)

                    # If not last step in animation, add a cursor to the line
                    is_last_part = (count + 1 == num_parts)
                    is_last_letter = (len(typewriter) == len(part.text))
                    if not (is_last_part and is_last_letter):
                        output_parts.append( CodePart(Token, '\u2588') )

                    row_line = CodeLine(output_parts, self.code.lexer)
                    step = steplib.ReplaceRows(self.code_box, replace_pos,
                        row_line)
                    steps.append(step)

                    steps.append(steplib.Sleep(self.delay_until_next_letter))

        return steps


    def steps(self):
        steps = []

        lines = parse_source(self.code.source, self.code.lexer)
        for count, line in enumerate(lines):
            if self.position == 0:
                # Append to end
                line_steps = self._line_to_steps(line, 0, -1)
            else:
                # Append to position
                spot = self.position + count
                line_steps = self._line_to_steps(line, spot, spot)

            steps.extend(line_steps)

        return steps


class AppendTypewriter(TypewriterStep):
    """Adds the content of a :class:`purdy.content.Code` object to a
    :class:`purdy.ui.CodeBox` using the typewriter animation.

    :param code_box: the :class:`purdy.ui.CodeBox` instance to append code
                     into

    :param code: a :class:`purdy.content.Code` object containing the source
                 code to insert.
    """

    def __init__(self, code_box, code):
        self.code_box = code_box
        self.code = code
        self.position = 0

    def __str__(self):
        content = condense(self.code.source)
        return f'actions.AppendTypewriter("{content}")'


class InsertTypewriter(TypewriterStep):
    """Inserts the contents of a :class:`purdy.content.Code` object at the
    given position using the typewriter animation.

    :param code_box: the :class:`purdy.ui.CodeBox` instance to append code
                     into

    :param position: line number to insert the code at. Position is 1-indexed.

    :param code: a :class:`purdy.content.Code` object containing the source
                 code to insert.
    """
    def __init__(self, code_box, position, code):
        if position < 0:
            raise AttributeError(
                'Negative indicies are not supported for this action')

        self.code_box = code_box
        self.code = code
        self.position = position

    def __str__(self):
        content = condense(self.code.source)
        return f'actions.InsertTypewriter({self.position}, "{content}")'


class SuffixTypewriter(TypewriterBase):
    """Adds the provided text to the end of an existing line in a
    :class:`purdy.ui.CodeBox` using a typewriter animation.

    :param code_box: the :class:`purdy.ui.CodeBox` instance to append code
                     into

    :param position: line number to insert the code at. Position is 1-indexed.
                     Negative indicies are supported.

    :param source: a string to be appended to the given line
    """
    def __init__(self, code_box, position, source):
        self.code_box = code_box
        self.position = position
        self.source = source

    def __str__(self):
        return f'actions.SuffixTypewriter({self.position}, "{self.source}")'

    def steps(self):
        steps = []
        for count, letter in enumerate(self.source):
            cursor = False
            if count + 1 != len(self.source):
                cursor = True

            steps.extend([
                steplib.SuffixRow(self.code_box, self.position, letter, cursor),
                steplib.Sleep(self.delay_until_next_letter),
            ])

        return steps

# =============================================================================
# Presentation Actions
# =============================================================================

class Highlight:
    """Cause one or more lines of code to have highlighting turned on or off

    :param code_box: :class:`purdy.ui.CodeBox` to perform on

    :param spec: either a string containing comma separated and/or hyphen
                 separated integers (e.g. "1,3,7-9") or a list of integers
                 specifying the lines in the code box to highlight. Line
                 numbers are 1-indexed

    :param highlight_on: True to turn highligthing on, False to turn it off
    """
    def __init__(self, code_box, spec, highlight_on):
        self.code_box = code_box
        self.highlight_on = highlight_on

        if isinstance(spec, str):
            self.numbers = range_set_to_list(spec)
        else:
            self.numbers = spec

    def __str__(self):
        return f'actions.Highlight({self.numbers}, {self.highlight_on})'

    def steps(self):
        return [steplib.HighlightLines(self.code_box, self.numbers,
            self.highlight_on) ]


class HighlightChain:
    """A common pattern with highlighting lines is to turn a highlight on for
    some set of lines, then turn it off and turn it on for more lines. This is
    a convenience wrapper to the :class:`Highlight` action, turning items on
    and off in series.

    :param code_box: :class:`purdy.ui.CodeBox` to perform series of highlight on

    :param spec_list: a list of highlight specs (see :class:`Highlight` for
                      details on a spec)
    """
    def __init__(self, code_box, spec_list):
        self.code_box = code_box
        self.spec_list = spec_list

    def __str__(self):
        return f'actions.HighlightChain({self.spec_list})'

    def steps(self):
        all_steps = []
        previous_numbers = None
        for spec in self.spec_list:
            if isinstance(spec, str):
                numbers = range_set_to_list(spec)
            else:
                numbers = spec

            if previous_numbers:
                all_steps.append( steplib.HighlightLines(self.code_box,
                    previous_numbers, False) )

            all_steps.extend([
                steplib.HighlightLines(self.code_box, numbers, True),
                steplib.CellEnd()
            ])

            previous_numbers = numbers

        return all_steps


class Fold:
    """Folds code by replacing one or more lines with a vertical elipses
    symbol.


    :param code_box: the :class:`purdy.ui.CodeBox` instance to modify

    :param position: line number to begin the fold at. Position is 1-indexed.

    :param end: line number to finish the folding at, inclusive. A value of -1
                can be used to fold to the end of the box. Defaults to -1.
    """
    def __init__(self, code_box, position, end=-1):
        self.code_box = code_box
        self.position = position
        self.end = end
        if end != -1 and end < position:
            raise AttributeError((f'Ending line number ({end}) must be '
                f'larger than starting number ({position})'))

    def __str__(self):
        return f'actions.Fold({self.position}, {self.end})'

    def steps(self):
        return [steplib.FoldLines(self.code_box, self.position, self.end)]

# =============================================================================
# Control Actions
# =============================================================================

class Wait:
    """Causes the animations to wait for a `right arrow` key press before
    continuing.
    """
    def __str__(self):
        return 'actions.Wait()'

    def steps(self):
        return [steplib.CellEnd(), ]


class StopMovie:
    """Causes the presentation :class:`purdy.ui.Screen` to exit movie mode"""
    def __str__(self):
        return 'actions.StopMovie()'

    def steps(self):
        return [steplib.StopMovie(), ]


class Transition:
    """Replaces the contents of a :class:`purdy.ui.CodeBox` with new content,
    doing a wipe animation from top to bottom. Only one of `code` or
    `code_box_to_copy` should be given, both can be blank to transition to an
    empty screen.

    :param code_box: the :class:`purdy.ui.CodeBox` instance to perform the
                     transition on

    :param code: a :class:`purdy.content.Code` object containing the source
                 code replacing the existing content. Should not be used at
                 the same time as code_box_to_copy parameter.

    :param code_box_to_copy: a code box containing rendered code to copy into
                             this one to display. This is typically a
                             :class:`VirtualCodeBox`. Should not be used at
                             the same time as code parameter.
    """
    def __init__(self, code_box, code=None, code_box_to_copy=None):
        self.code_box = code_box
        self.code = code
        self.code_box_to_copy = code_box_to_copy

    def __str__(self):
        code_content = ''
        if self.code:
            code_content = condense(self.code.source)

        copy_content = ''
        if self.code_box_to_copy:
            lines = [str(line) for line in self.code_box_to_copy.listing.lines]
            copy_content = condense('\n'.join(lines))

        return f'actions.Transition("{code_content}", "{copy_content}")'

    def steps(self):
        return [steplib.Transition(self.code_box, self.code,
            self.code_box_to_copy), ]


class Sleep:
    """Causes animations to pause for the given amount of time. Note that this
    action happens within a cell, so is considered part of the group of
    animation steps done together. For example if Append + Sleep + Append is
    part of the same cell it is all done/undone together.

    :param time: Either the amount of time to sleep in seconds (ints and 
                 floats supported), or a tuple containing a pair of times
                 specifying the range of a random value to sleep.

    """
    def __init__(self, time):
        self.time = time

    def __str__(self):
        return 'actions.Sleep'

    def steps(self):
        return [steplib.Sleep(self.time), ]


class RunFunction:
    """Calls the function passed in, allowing the execution of code during the
    playing of actions.

    :param fn: function to be called
    :param undo: function to be called when this Action is undone, can be None
    :param \*args, \*\*kwargs: any remaining arguments are passed to the
                            functions when they are called
    """
    def __init__(self, fn, undo, *args, **kwargs):
        self.fn = fn
        self.undo = undo
        self.fn_args = args
        self.fn_kwargs = kwargs

    def __str__(self):
        return 'actions.RunFunction'

    def steps(self):
        return [steplib.RunFunction(self.fn, self.undo, *self.fn_args,
            **self.fn_kwargs), ]


class Section:
    """Marker for the beginning of a section. In the TUI you can skip to the
    next section marker using "S"."""

    def __str__(self):
        return 'actions.Section()'

    def steps(self):
        return [steplib.SectionBreak(), ]



from unittest import TestCase

from pygments.token import Token

from purdy.parser import (PurdyLexer, BlankCodeLine, CodeLine, CodePart, 
    token_is_a, token_ancestor)

# =============================================================================

PYCON_SOURCE = """\
>>> x = 1
"""

PY_SCRIPT = """\
#!/usr/bin/env python

print('hello')
"""

BASHCON_SOURCE = """\
$ echo "hello"
hello
"""

class TestParser(TestCase):
    def test_lexer_container(self):
        #--- Test the names property
        names = list(PurdyLexer.names())
        self.assertEqual(3, len(names))
        names = set(names)

        expected = set(['py3', 'con', 'bash'])
        self.assertEqual(expected, names)

        #--- Test the choices property
        choices = PurdyLexer.choices().split(',')
        self.assertEqual(3, len(choices))

        self.assertIn('"py3" (Python 3 Source)', PurdyLexer.choices())
        self.assertIn('"con" (Python 3 Console)', PurdyLexer.choices())
        self.assertIn('"bash" (Bash Console)', PurdyLexer.choices())

        #--- Test lexer detection
        lexer = PurdyLexer.factory_from_source(PYCON_SOURCE)
        self.assertEqual('con', lexer.name)

        lexer = PurdyLexer.factory_from_source(PY_SCRIPT)
        self.assertEqual('py3', lexer.name)

        lexer = PurdyLexer.factory_from_source(BASHCON_SOURCE)
        self.assertEqual('bash', lexer.name)

    def test_tokens(self):

        #--- Test token_is_a()
        self.assertTrue( token_is_a(Token, Token) )
        self.assertTrue( token_is_a(Token.Name.Function, Token.Name) )
        self.assertFalse( token_is_a(Token.Name.Function, Token.Keyword) )

        #--- Test token_ancestor()
        token = token_ancestor(Token.Name.Function, [Token.Name.Function])
        self.assertEqual(Token.Name.Function, token)

        token = token_ancestor(Token.Name.Function, [Token.Name])
        self.assertEqual(Token.Name, token)

        token = token_ancestor(Token.Text, [Token.Name])
        self.assertEqual(Token, token)

    def test_misc(self):
        blank = BlankCodeLine()
        self.assertEqual('', blank.render_line(None))

        lexer = PurdyLexer.factory_from_name('py3')
        line = CodeLine([CodePart(Token.Text, 'foo'), ], lexer, line_number=10)

        expected = 'CodeLine(" 10 foo")'
        self.assertEqual(expected, line.__repr__())



from unittest import TestCase

from pygments.token import Token

from purdy.parser import CodePart, CodeLine, PurdyLexer

# =============================================================================

py3_lexer = PurdyLexer.factory_from_name('py3')
bash_lexer = PurdyLexer.factory_from_name('bash')

PY_CODE_LINES = [
    CodeLine([
        CodePart(Token.Comment.Single, '# Sample Code'),
    ], py3_lexer),
    CodeLine([
        CodePart(Token.Text, ''),
    ], py3_lexer),
    CodeLine([
        CodePart(Token.Keyword, 'def'),
        CodePart(Token.Text, ' '),
        CodePart(Token.Name.Function, 'foo'),
        CodePart(Token.Punctuation, '('),
        CodePart(Token.Punctuation, ')'),
        CodePart(Token.Punctuation, ':'),
    ], py3_lexer),
    CodeLine([
        CodePart(Token.Text, '    '),
        CodePart(Token.Literal.String.Doc, '"""Multi-line'),
    ], py3_lexer),
    CodeLine([
        CodePart(Token.Literal.String.Doc, '    Doc string'),
    ], py3_lexer),
    CodeLine([
        CodePart(Token.Literal.String.Doc, '    """'),
    ], py3_lexer),
    CodeLine([
        CodePart(Token.Text, '    '),
        CodePart(Token.Name, 'bar'),
        CodePart(Token.Text, ' '),
        CodePart(Token.Operator, '='),
        CodePart(Token.Text, ' '),
        CodePart(Token.Literal.String.Single, "'"),
        CodePart(Token.Literal.String.Single, 'Thing'),
        CodePart(Token.Literal.String.Single, '\\'),
        CodePart(Token.Literal.String.Single, 'Stuff'),
        CodePart(Token.Literal.String.Single, "'"),
    ], py3_lexer),
]

BASH_CODE_LINES = [
    CodeLine([
        CodePart(Token.Generic.Prompt, '$ '),
        CodePart(Token.Name.Builtin, 'echo'),
        CodePart(Token.Text, ' '),
        CodePart(Token.Literal.String.Double, '"hello\\nthere"'),
    ], bash_lexer),
    CodeLine([
        CodePart(Token.Generic.Output, 'hello'),
    ], bash_lexer),
    CodeLine([
        CodePart(Token.Generic.Output, 'there'),
    ], bash_lexer),
    CodeLine([
        CodePart(Token.Generic.Prompt, '$ '),
    ], bash_lexer),
]

# =============================================================================

def data_lines_to_source(data):
    text = []
    for line in data:
        words = []
        for part in line.parts:
            words.append(part.text)

        text.append(''.join(words))

    return '\n'.join(text)

# =============================================================================

class PurdyContentTest(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.py_source = data_lines_to_source(PY_CODE_LINES)
        cls.bash_source = data_lines_to_source(BASH_CODE_LINES)



"""
Content
=======

Reperesntations of source code are found in this module.
"""
import ast, asttokens, os
from copy import deepcopy

from purdy.colour import COLOURIZERS
from purdy.parser import FoldedCodeLine, parse_source

#import logging
#logging.basicConfig(filename='debug.log', level=logging.DEBUG)
#logger = logging.getLogger()

# =============================================================================

class Code:
    """Represents source code from the user.

    :param filename: name of a file to read for content. If both this and 
                     `text` is given, `filename` is used first

    :param text: a string containing code

    :param lexer_name: name of lexer to use to tokenize the code, defaults to 
                       'detect', attempting to auto detect the type of
                       content. See :class:`purdy.parser.LexerContainer` for a
                       list of available lexers.

    :param purdy_lexer: if lexer_name is "custom" this parameter is expected
                         to contain a purdy.parser.PurdyLexer object.
    """

    def __init__(self, filename='', text='', lexer_name='detect', 
            purdy_lexer=None):
        self.source = ''

        if filename:
            filename = os.path.abspath(filename)
            with open(filename) as f:
                self.source += f.read()

        if text:
            self.source += text

        self.source = self.source.rstrip('\n')

        from purdy.parser import PurdyLexer
        if lexer_name == 'detect':
            self.lexer = PurdyLexer.factory_from_source(self.source)
        elif lexer_name == 'custom':
            self.lexer = purdy_lexer
        else:
            self.lexer = PurdyLexer.factory_from_name(lexer_name)

    def remove_double_blanks(self, trim_whitespace=True):
        """Removes the second of two blanks in a row. If trim_whitespace is
        True (default) a line with only whitespace is considered blank,
        otherwise it only looks for \\n"""
        lines = self.source.split('\n')
        output = []

        previous = 'asdf'
        for line in lines:
            if trim_whitespace \
                    and previous.strip() == '' and line.strip() == '':
                continue
            else:
                if previous == '' and line == '':
                    continue

            output.append(line)
            previous = line

        self.source = '\n'.join(output)
        return self

    def remove_lines(self, line_no, count=1):
        """Removes one or more lines from the source listing.

        :param line_no: number of the line to remove, 1-indexed
        :param count: number of lines to remove, defaults to 1
        """
        lines = self.source.split('\n')
        start = line_no - 1
        end = start + count
        del lines[start:end]
        self.source = '\n'.join(lines)
        return self

    def replace_line(self, line_no, content):
        """Replaces the given line with new content

        :param line_no: number of the line to replace, 1-indexed
        :param content: content to replace it with 
        """
        lines = self.source.split('\n')
        lines[line_no - 1] = content
        self.source = '\n'.join(lines)
        return self

    def inline_replace(self, line_no, pos, content):
        """Replaces the contents of a line starting at pos with the new
        content

        :param line_no: number of the line to replace, 1-indexed
        :param pos: position number to start the replacement at, 1-indexed
        :param content: content to replace with 
        """
        lines = self.source.split('\n')
        lines[line_no - 1] = lines[line_no - 1][:pos - 1]
        lines[line_no - 1] += content
        self.source = '\n'.join(lines)
        return self

    def insert_line(self, line_no, content):
        """Inserts the given line into the source, pushing the content down
        from the given line number

        :param line_no: number of the line to insert at, 1-indexed
        :param content: content to insert
        """
        lines = self.source.split('\n')
        lines.insert(line_no - 1, content)
        self.source = '\n'.join(lines)
        return self

    def _descend_ast(self, atok, parent, local_name, name_parts):
        for node in ast.iter_child_nodes(parent):
            if node.__class__ in [ast.FunctionDef, ast.ClassDef] \
                    and node.name == local_name:
                if len(name_parts) == 0:
                    # Found the node, return the code
                    return atok.get_text(node) + "\n"

                # Found a function or class, but looking for a child of it
                return self._descend_ast(atok, node, name_parts[0],
                    name_parts[1:])

            if node.__class__ == ast.Assign and \
                    node.first_token.string == local_name and \
                    len(name_parts) == 0:
                return atok.get_text(node) + "\n"

        return ''

    def python_portion(self, name):
        """Treates the source in this object as Python and then finds either
        the named function, class, or assigned variable and replaces the 
        source with only the found item. 

        .. warning:: If the named item is not found your source will be empty!

        :param name: dot notated name of a function or class. Examples:
                     `Foo.bar` would find the `bar` method of class `Foo` or
                     an inner function named `bar` in a function named `Foo`
        """
        atok = asttokens.ASTTokens(self.source, parse=True)
        name_parts = name.split('.')
        self.source = self._descend_ast(atok, atok.tree, name_parts[0],
            name_parts[1:])
        return self

    def left_justify(self):
        """Removes a consistent amount of leading whitespace from the front of
        each line so that at least one line is left-justified.

        .. warning:: will not work with mixed tabs and spaces
        """
        lines = self.source.split('\n')
        leads = [len(line) - len(line.lstrip()) for line in lines if \
            len(line.strip())]
        if not leads:
            # only blank lines, do nothing
            return

        min_lead = min(leads)
        output = []
        for line in lines:
            if len(line.lstrip()):
                output.append(line[min_lead:])
            else:
                output.append(line)

        self.source = '\n'.join(output)
        return self

    def fold_lines(self, start, end):
        """Call this method to replace one or more lines with a vertical
        elipses, i.e. a fold of the code. 

        :param start: line number of the listing to start code folding on. 
                      1-indexed.
        :param end: line number to fold until, inclusive (1-indexed).
        """
        lines = self.source.split('\n')

        self.source = '\n'.join(lines)

        size = end - start + 1
        lines[start - 1] = '⋮'

        if size > 1:
            del lines[start:end]

        self.source = '\n'.join(lines)

        return self

    def subset(self, start, end):
        """Returns a new Code object containing just the subset of lines
        identified in this call

        :param start: line number of the listing to start the subset at.
                      1-indexed
        :param end: line number to finish the subset on, inclusive.  1-indexed.

        :returns: Code object
        """
        code = deepcopy(self)
        if end < start:
            raise AttributeError('Value of end must be bigger than start')

        lines = self.source.split('\n')
        lines = lines[start - 1:end]
        code.source = '\n'.join(lines)
        return code

# -----------------------------------------------------------------------------

class RenderHook:
    def line_inserted(self, listing, position, line):
        pass

    def line_removed(self, listing, position):
        pass

    def line_changed(self, listing, position, line):
        pass

    def clear(self):
        pass


class Listing:
    """A container for a set of :class:`CodeLine` objects."""

    def __init__(self, code=None, starting_line_number=-1):
        """Constructor

        :param code: a :class:`Code` or list of :class:`Code` objects to
                     populate this display with
        :param starting_line_number: If a non-negative integer, this value is
                                     the starting value of line numbers
                                     displayed with the code. A value of -1,
                                     the default, turns line numbering off
        """
        self.starting_line_number = starting_line_number

        self.colourizer = COLOURIZERS['plain']
        self.render_hook = RenderHook()
        self.lines = []

        if code:
            if isinstance(code, Code):
                self.append_code(code)
            else:
                for item in code:
                    self.append_code(item)


    def set_display(self, mode='plain', render_hook=None):
        """Code can be displayed using a variety of methods, from colourized
        text to the urwid TUI client. This method sets the options to control
        how rendering is done.

        :param mode: a key from :attr:`purdy.colour.COLOURIZERS` indicating
                     which colourizer routine is used to colour the source in
                     this box. Defaults to 'plain'.
        :param render_hook: an optional hook the box calls when actions are
                            done on the box. Used by the TUI client so that
                            when something changes in the it is reflected in
                            the TUI.
        """
        self.colourizer = COLOURIZERS[mode]

        if render_hook:
            self.render_hook = render_hook

    #--- Add/Remove Line Methods
    def positive_position(self, position):
        original = position
        size = len(self.lines)

        if position < 0:
            position = size + position + 1

        # check within bounds
        if position <= 0 or position > size:
            if original == 0:
                raise IndexError('position is 1-indexed, 0 is invalid')

            if original < 0:
                raise IndexError( (f'position {original} translated to '
                    f'{position}, is outside range 1-{size}') )

            raise IndexError(f'position {position} outside range 1-{size}')

        return position

    def append_code(self, code):
        """Parses contents of a :class:`Code` object and appends it to this
        box. Has a side effect of changing the listing's colour palette to
        that of the code object sent in."""
        lines = parse_source(code.source, code.lexer)
        self.insert_lines(0, lines)

    def insert_lines(self, position, lines):
        """Inserts a line at the given position. Pushes content down, so
        a line inserted at position 1 becomes the new first item.

        :param position: 1-indexed position in the listing. Supports negative
                         indexing and a special value of 0 to indicate
                         appending to the list
        :param lines: lines to insert
        """
        # loop through lines and insert them, assigning a new line number if
        # needed
        if position == 0:
            # special case, insert becomes append
            position = len(self.lines) + 1
        else:
            position = self.positive_position(position)

        new_lines = deepcopy(lines)

        for count, line in enumerate(new_lines):
            if self.starting_line_number > -1:
                line.line_number = position + count + \
                    self.starting_line_number - 1

            self.lines.insert(count + position - 1, line)
            self.render_hook.line_inserted(self, position + count, line)

        if self.starting_line_number > -1:
            self.reset_line_numbers()

    def replace_line(self, position, line):
        """Replaces the line at the given position with the given
        :class:`CodeLine` object"""
        position = self.positive_position(position)

        new_line = deepcopy(line)
        if self.starting_line_number != -1:
            new_line.line_number = position + self.starting_line_number - 1

        self.lines[position - 1] = new_line
        self.render_hook.line_changed(self, position, new_line)

    def remove_lines(self, position, size=1):
        position = self.positive_position(position)

        for x in range(1, size + 1):
            del self.lines[position - 1]

            self.render_hook.line_removed(self, position)

        if self.starting_line_number > -1:
            self.reset_line_numbers()

    def clear(self):
        self.lines = []
        self.render_hook.clear()

    def copy_lines(self, position, size=1):
        """Returns a copy of the lines for the given position and size

        :param position: 1-index position in the listing, supports negative
                         indexing
        :param size: number of lines to return, defaults to 1
        """
        position = self.positive_position(position)
        result = []

        for count in range(0, size):
            result.append( deepcopy(self.lines[position - 1 + count]) )

        return result

    def fold_lines(self, start, end):
        """Call this method to replace one or more lines with a vertical
        elipses, i.e. a fold of the code. 

        :param start: line number of the listing to start code folding on. 
                      1-indexed.
        :param end: line number to fold until, inclusive (1-indexed).
        """
        size = end - start + 1

        replace_line = FoldedCodeLine(size)
        self.lines[start - 1] = replace_line
        self.render_hook.line_changed(self, start, replace_line)

        if size > 1:
            #remove rest of lines in fold, line numbers will be fixed when
            # remove_lines() calls reset_line_numbers()
            self.remove_lines(start + 1, size - 1)

    #--- Style Methods

    def set_highlight(self, highlight, start=1, end=-1):
        """Turns highlighting on for the lines of code

        :param highlight: boolean value, True for turning highlighting on
        :param start: line number to start highlighting on (1-indexed), 
                      defaults to 1 
        :param end: line number to highlight until, inclusive (1-indexed),
                    a value of -1 indicates to change the highlight from start
                    until the end of the list. Defaults to -1
        """
        first = start - 1
        last = end
        if end == -1:
            last = len(self.lines)

        for count, line in enumerate(self.lines[first:last]):
            index = start + count
            line.highlight = highlight
            self.render_hook.line_changed(self, index, line)

    def reset_line_numbers(self):
        """Reset the line numbers."""
        # Because of code folding possibly advancing the numbering, have to
        # check the whole listing
        current_position = 1
        line_number = self.starting_line_number

        if len(self.lines) > 0 and isinstance(self.lines[0], FoldedCodeLine):
            line_number = self.starting_line_number + self.lines[0].size

        for line in self.lines:
            if not isinstance(line, FoldedCodeLine) and \
                    line.line_number != line_number:
                # changing the line number, trigger the hook
                line.line_number = line_number
                self.render_hook.line_changed(self, current_position, line)

            current_position += 1
            if isinstance(line, FoldedCodeLine):
                line_number += line.size
            else:
                line_number += 1

    #--- Export methods
    def content(self):
        """Returns a list of colourized strings. Uses the colourizer set with
        :funct:`Listing.set_display`.
        """
        result = []
        for line in self.lines:
            result.append( self.render_line(line) )

        return result

    def render_line(self, line):
        return line.render_line(self.colourizer)



"""
UI (purdy.ui.py)
================

This module is the entry point for the code viewers. It is a lightweight proxy
to implementation of a screen. Screen implementations are found in
:mod:`purdy.iscreen`.  All programs using the purdy library need to create a
:class:`Screen` object or one of its children. The factory in this module
determines which actual implementation is loaded.
"""
import argparse, sys

from purdy.animation.manager import AnimationManager
from purdy.animation.cell import group_steps_into_cells
from purdy.cmd import background_arg, max_height_arg
from purdy.content import Listing
from purdy.settings import settings as default_settings

#import logging
#logging.basicConfig(filename='debug.log', level=logging.DEBUG)
#logger = logging.getLogger()

# =============================================================================

DESCRIPTION = """You can alter the behaviour of any script calling
Screen.run() by passing in command line arguments."""

class Factory:
    name = 'tui'
    iscreen = None

    @classmethod
    def get_iscreen_class(cls):
        if cls.name == 'tui':
            from purdy.iscreen.tui.iscreen import TUIScreen
            return TUIScreen
        elif cls.name == 'virtual':
            from purdy.iscreen.virtual.iscreen import VirtualScreen
            return VirtualScreen
        elif cls.name == 'exercise':
            from purdy.iscreen.exercise.iscreen import ExerciseScreen
            return ExerciseScreen

# =============================================================================
# Screen
# =============================================================================

class VirtualCodeWidget:
    """Render hook implementation used with the :class:`VirtualCodeBox`.
    """
    def __init__(self):
        self.lines = []

    #--- RenderHook methods
    def line_inserted(self, listing, position, line):
        content = listing.render_line(line)
        index = position - 1
        self.lines.insert(index, content)

    def line_removed(self, listing, position):
        del self.lines[position - 1]

    def line_changed(self, listing, position, line):
        content = listing.render_line(line)
        index = position - 1
        self.lines[index] = content

    def clear(self):
        self.lines = []


class VirtualCodeBox:
    """Specifies a box to contain code. This implementation is not meant to be
    associated with a :class:`Screen`. Used for "rendering" off screen,
    manipulating code either for export or copying to a concrete code box
    later

    :param starting_line_number: -1 means no line numbers, anything larger will 
                                 be the first line number displayed.  Defaults 
                                 to -1
    """
    def __init__(self, starting_line_number=-1, display_mode='plain'):
        self.starting_line_number = starting_line_number
        self.listing = Listing(starting_line_number=starting_line_number)
        self.widget = VirtualCodeWidget()

        self.listing.set_display(display_mode, self.widget)

    def perform_actions(self, actions):
        """Performs the list of actions on the code in the box. Done separtely
        from the main event loop so it has no interactions like Wait,
        typewriters, etc. Everything is done in fast-forward mode.

        :param actions: list of actions to perfrom on the code in this box
        """
        for action in actions:
            for step in action.steps():
                step.render_step()

# -----------------------------------------------------------------------------

class CodeBox:
    """Specifies a box to contain code. :class:`Screen` uses these to
    determine widget layout, subsequent actions are done within the context of 
    this box. When :func:`CodeBox.build` is called by a :class:`Screen` class
    a widget is built and this box is added to the screen.

    :param starting_line_number: -1 means no line numbers, anything larger will 
                                 be the first line number displayed.  Defaults 
                                 to -1

    :param auto_scroll: When True, :class:`purdy.widgets.CodeBox` created by 
                        this specification automatically scrolls to newly added 
                        content.  Defaults to True.

    :param height: Number of lines the row containing this box should be.  A 
                   value of 0 indicates automatic spacing.  Defaults to 0.

    :param compact: if False, the dividing line between this box and the next 
                    has a 1-line empty boundary. Parameter is ignored if there 
                    is no item after this one in the rows=[] listing. Defaults 
                    to False
    """
    def __init__(self, starting_line_number=-1, auto_scroll=True, height=0, 
            compact=False):
        self.starting_line_number = starting_line_number
        self.auto_scroll = auto_scroll
        self.height = height
        self.compact = compact
        self.listing = Listing(starting_line_number=starting_line_number)

    def setup(self, screen):
        self.screen = screen
        self.screen.code_boxes.append(self)

    def build(self, screen):
        self.screen.iscreen.add_code_box(self)


class TwinCodeBox:
    """Specifies two side-by-side CodeBoxes. The contained CodeBoxes can be
    either accessed as twin.left and twin.right, or twin[0] and twin[1]. When
    :func:`TwinCodeBox.build` is called, the widgets are created and added to 
    the :attr:`Screen.code_boxes` list sequentially, i.e. a single
    TwinCodeBox results in two CodeBox items in Screen's list.

    :param left_starting_line_number:
    :param right_starting_line_number: -1 to turn line numbers off, any higher 
                                       value is the first number to display. 
                                       Defaults to -1

    :param left_auto_scroll:
    :param right_auto_scroll: True to specify that scrolling happens
                              automatically for the left and right boxes
                              created by this spec. Defaults to True.

    :param left_weight:
    :param right_weight: relative weights for the widths of the columns, if the 
                          values are the same then the columns are the same
                          width, otherwise the widths are formed based on the
                          ratio of left:right. Example: left_weight=2,
                          right_weight=1 means the left side will be twice the
                          width of the right side. Both values default to 1.

    :param height: number of lines for the row this set of boxes is in.  The 
                   default of 0 specifies automatic height

    :param compact: if False, the dividing line between this box and the next 
                    has a 1-line empty boundary. Parameter is ignored if there
                    is no item after this one in the rows=[] listing. Defaults
                    to False
    """
    def __init__(self, left_starting_line_number=-1, left_auto_scroll=True, 
            left_weight=1, right_starting_line_number=-1, 
            right_auto_scroll=True, right_weight=1, height=0, compact=False):

        self.left_starting_line_number = left_starting_line_number
        self.left_auto_scroll = left_auto_scroll
        self.left_weight = left_weight
        self.right_starting_line_number = right_starting_line_number
        self.right_auto_scroll = right_auto_scroll
        self.right_weight = right_weight
        self.height = height
        self.compact = compact

    def setup(self, screen):
        class Side:
            pass

        self.screen = screen
        self.left = Side()
        self.left.code_box = CodeBox(self.left_starting_line_number,
            self.left_auto_scroll)
        self.left.code_box.setup(screen)

        self.right = Side()
        self.right.code_box = CodeBox(self.right_starting_line_number,
            self.right_auto_scroll)
        self.right.code_box.setup(screen)

    def build(self, screen):
        self.screen.iscreen.add_twin_code_box(self)


class Screen:
    """Represents the main UI window for the TUI application. The layout is
    specified by passing in one or more :class:`CodeBox` or
    :class:`TwinCodeBox` objects to the constructor. Each box will have a
    corresponding :class:`purdy.widgets.CodeWidget` inside of the UI for
    displaying code.

    :param settings: a settings dictionary object. Defaults to `None` which 
                     uses the default settings dictionary: 
                     :attr:`settings.settings`

    :param rows: a list containing one or more :class:`CodeBox` or 
                 :class:`TwinCodeBox` definitions, to specify the layout of 
                 the screen

    :param max_height: maximum display height in TUI mode, defaults to 0,
                       meaning no max

    Example:

    .. code-block:: python

        from purdy.actions import AppendAll
        from purdy.content import Code
        from purdy.ui import Screen, CodeBox, TwinCodeBox

        screen = Screen(rows=[TwinCodeBox(height=8),
                CodeBox(auto_scroll=False)])

        c1 = Code(contents_filename='c1.py', starting_line_number=1)
        c2 = Code(contents_filename='c2.py')
        c3 = Code(contents_filename='c3.py')

        actions = [
            AppendAll(screen.code_boxes[0], c1),
            AppendAll(screen.code_boxes[1], c2),
            AppendAll(screen.code_boxes[2], c3),
        ]

        screen.run(actions)

    The above would produce a screen with two rows, the first row having two 
    :class:`purdy.widgets.CodeWidget` objects side by side, the second having a
    single one. The top left box would have line numbers turned on, both top
    boxes are 8 lines tall, and the bottom box has auto scrolling turned off.

    The screen would be divided like this:

    +-----------+-----------+
    |           |           |
    |           |           |
    +-----------+-----------+
    |                       |
    |                       |
    |                       |
    |                       |
    +-----------------------+

    Command Line Parameters
    -----------------------

    Unless "deactivate_args" is set to True in :mod:`purdy.settings` (False by
    default), Screen will also parse command line arguments. This allows
    scripts calling the library to change their behaviour with switches. 

    Supported switches are:

    * **--debugsteps** Print out the animation steps, grouped by Cell and exit
    * **--export** Print out the results of the actions
    * **--exportrtf** Print out the results of the actions in RTF format

    """
    def __init__(self, settings=None, rows=[], max_height=0):
        self.rows = rows
        self.max_height = max_height
        self.code_boxes = []
        self.concrete_boxes = []

        self.settings = settings
        if settings is None:
            self.settings = default_settings

        self.movie_mode = self.settings['movie_mode']
        if self.movie_mode != -1:
            self.movie_mode = float(self.movie_mode) / 1000.0

        self.animation_manager = AnimationManager(self)

        for row in self.rows:
            row.setup(self)

    def _get_args(self):
        class Dummy:
            pass

        # Screen can be influenced by command line arguments
        parser = argparse.ArgumentParser(description=DESCRIPTION)
        parser.add_argument('--debugsteps', '--ds', action='store_true', 
            help='Print out the animation steps, grouped by Cell and exit')

        parser.add_argument('--export', action='store_true', 
            help='Print out the results of the actions')

        parser.add_argument('--exportrtf', action='store_true', 
            help='Print out the results of the actions in RTF format')

        background_arg(parser)
        max_height_arg(parser)

        if self.settings['deactivate_args']:
            # Ignore whatever was passed in on command line because of the
            # settings
            self.args = parser.parse_args([])
        else:
            self.args = parser.parse_args()
            if self.args.maxheight:
                self.max_height = self.args.maxheight

    def load_actions(self, actions):
        steps = []
        for (index, action) in enumerate(actions):
            try:
                steps.extend( action.steps() )
            except:
                tb = sys.exc_info()[2]
                raise RuntimeError( ('An exception occurred while loading '
                    f'action #{index+1}: {action.__class__.__name__}'
                    )).with_traceback(tb)

        cells = group_steps_into_cells(steps)
        self.animation_manager.register(cells)

    def set_alarm(self, handler, when):
        return self.iscreen.set_alarm(handler, when)

    def remove_alarm(self, alarm_handle):
        self.iscreen.remove_alarm(alarm_handle)

    def run(self, actions):
        """Calls the main display event loop. Does not return until the UI
        exits."""
        #logger.debug(55*'=')
        self._get_args()
        self.actions = actions
 
        if self.args.export or self.args.exportrtf:
            Factory.name = 'virtual'

        # Create concrete instances of iscreen an boxes
        klass = Factory.get_iscreen_class()
        self.iscreen = klass(self)
        for row in self.rows:
            row.build(self)

        # Process the actions
        self.load_actions(actions)

        if self.args.debugsteps:
            for cell in self.animation_manager.cells:
                print(f'{cell.__class__.__name__}')
                for step in cell.steps:
                    print('   step', step)
            exit()

        self.iscreen.run()

# =============================================================================
# Other Screens
# =============================================================================

class SplitScreen(Screen):
    """Convenience implementation of :class:`Screen` that supports two 
    :class:`CodeBox` instances, stacked vertically and separated by a dividing
    line. The code boxes are :attr:`SplitScreen.top` and
    :attr:`SplitScreen.bottom`. 

    :param settings: a settings dictionary object. Defaults to `None` which 
                     uses the default settings dictionary: 
                     :attr:`settings.settings`

    :param top_starting_line_number: starting line number for the top 
                                     code box

    :param top_auto_scroll: When True, the top :class:`ui.CodeBox` 
                            automatically scrolls to newly added content.
                            Defaults to True.

    :param bottom_starting_line_number: starting line number for the 
                                        bottom code box

    :param bottom_auto_scroll: When True, the bottom :class:`ui.CodeBox` 
                            automatically scrolls to newly added content.
                            Defaults to True.

    :param top_height: Number of lines the top box should be. A value of 0
                       indicates top and bottom should be the same size.
                       Defaults to 0.

    :param compact: True if for the dividing line between the top and bottom 
                    screens is to have no margin. Defaults to False

    :param max_height: maximum display height in TUI mode, defaults to 0,
                       meaning no max
    """
    def __init__(self, settings=None, top_starting_line_number=-1, 
            top_auto_scroll=True, bottom_starting_line_number=-1,
            bottom_auto_scroll=True, top_height=0, compact=False, 
            max_height=0):
        self.top = CodeBox(top_starting_line_number, top_auto_scroll, 
            top_height, compact=compact)
        self.bottom = CodeBox(bottom_starting_line_number, bottom_auto_scroll)
        super().__init__(settings, rows=[self.top, self.bottom], 
            max_height=max_height)


class SimpleScreen(Screen):
    """Convenience implementation of :class:`Screen` that supports a single
    :class:`CodeBox`.  The code box is available as 
    :attr:`SimpleScreen.code_box`.

    :param settings: a settings dictionary object. Defaults to `None` which 
                     uses the default settings dictionary: 
                     :attr:`settings.settings`
    
    :param starting_line_number: starting line number for the created code
                                 box
    :param auto_scroll: When True, the class:`ui.CodeBox` automatically 
                        scrolls to newly added content.  Defaults to True.

    :param max_height: maximum display height in TUI mode, defaults to 0,
                       meaning no max
    """
    def __init__(self, settings=None, starting_line_number=-1, 
            auto_scroll=True, max_height=0):
        """Foo man chooo"""
        self.code_box = CodeBox(starting_line_number, auto_scroll)
        super().__init__(settings, [self.code_box], max_height=max_height)

    def thing(self):
        pass



import json
from pathlib import Path
from unittest import TestCase

from tests.capture_exercise import run_exercise


class TestExercises(TestCase):
    def test_exercises(self):
        exercises = ["all_console", "append", "bash_console", "builder", 
            "cols", "compact", "docstring", "exec", "fold", "highlight",
            "insert", "lines", "maxheight", "mixed", "mls_type",
            "movie_console", "remove", "replace", "section", "sleep", 
            "split_long", "src_change", "swipe", "tall", "triple",
            "type_console", ]

        exercise_data = Path(__file__).parent / "exercise_data"

        for exercise in exercises:
            # load stored expected results
            filename = exercise_data / f"{exercise}.json"
            with open(filename) as f:
                expected = json.load(f)

            # load the module and run the actions using the ExerciseScreen
            pages = run_exercise(exercise)

            try:
                self.assertEqual(expected, pages, 
                    msg=f"Compare failed for exercise: {exercise}")
            finally:
                pages.insert(0, f"******** Exercise: {exercise}")
                with open("last_exercise.json", "w") as f:
                    json.dump(pages, f, indent=2)
